from flask import Response, request, abort
from flask_restx import Namespace, Resource, fields
from flask_caching import Cache
import json, haplotyping, sqlite3, os

from haplotyping.service.kmer_kmc import Kmer as KmerKMC
from haplotyping.service.split import Split

def _make_cache_key(*args, **kwargs):
    #todo: find better solution    
    try:
        cacheKey = "%s_%s_%s_%s" % (args[0].__class__.__name__, str(request.path), str(request.args), str(namespace.payload))
    except:
        cacheKey = "%s_%s_%s" % (args[0].__class__.__name__, str(request.path), str(request.args))
    return cacheKey
    
namespace = Namespace("kmer", description="K-mer frequencies for a dataset", path="/kmer")
cache = Cache()
parser = namespace.parser()
            
def _getDataset(uid):
    db_connection = haplotyping.service.API.get_db_connection()
    db_connection.row_factory = sqlite3.Row
    cursor = db_connection.cursor()
    cursor.execute("SELECT `dataset`.`uid`, \
                    `dataset`.`type`, \
                    `dataset`.`location` AS `dataset_location`, \
                    `collection`.`location` AS `collection_location` \
                    FROM `dataset` \
                    LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                    WHERE `dataset`.`uid` = ? \
                    AND (`dataset`.`type` = 'kmer' OR `dataset`.`type` = 'split') \
                    AND NOT `collection`.`location` IS NULL",(str(uid),))  
    data = cursor.fetchone()
    return data

@namespace.route("/<uid>/distribution")
class KmerDistribution(Resource):
    
    @namespace.doc(description="Get k-mer frequency distribution from dataset defined by uid. Can only be used if a splitting k-mer database is available!")
    
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid):
        try:
            dsData = _getDataset(uid)
            if dsData:
                if not dsData["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                dsData["collection_location"],dsData["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                dsData["dataset_location"],"kmer.data.h5")
                location_split = os.path.abspath(location_split)
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_distribution(location_split)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))

@namespace.route("/<uid>/<kmer>")
class KmerSingle(Resource):
    
    dataset_kmers = parser.copy()
    dataset_kmers.add_argument("mismatches", type=int, required=False, location="values", 
                              help="maximum number of mismatches", choices=[0,1,2])
    
    @namespace.doc(description="Get k-mer frequencies from dataset defined by uid")
    @namespace.expect(dataset_kmers)
    @namespace.doc(params={"uid": "unique identifier dataset","kmer": "k-mer"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,kmer):
        mm = min(2,max(0,int(request.args.get("mismatches",0))))
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.kmc")
                else:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.kmc")
                location_kmc = os.path.abspath(location_kmc)
                if not os.path.isfile("{}.kmc_pre".format(location_kmc)):
                    abort(500,"kmc database not found")
                kmc_query_library = haplotyping.service.API.get_kmc_query_library()
                kmc_query_binary_location = haplotyping.service.API.get_kmc_query_binary_location()
                if kmc_query_library:
                    response = KmerKMC.kmc_library(kmc_query_library,location_kmc,[kmer],mm)
                    if not response:
                        abort(500,"no response using kmc query library for "+str(kmer))
                elif kmc_query_binary_location:
                    kmc_query_binary_location_query = os.path.join(kmc_query_binary_location,"kmc_query")
                    kmc_query_binary_location_analysis = os.path.join(kmc_query_binary_location,"kmc_analysis")
                    response = KmerKMC.kmc_binary_frequencies(kmc_query_binary_location_query,location_kmc,[kmer],mm)
                    if not response:
                        abort(500,"no response using kmc query binary for "+str(kmer))
                    response["info"] = KmerKMC.kmc_binary_info(kmc_query_binary_location_analysis,location_kmc)
                else:
                    abort(500,"no kmc binary or library configured")
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>/<kmer>/split")
class KmerSplit(Resource):
    
    dataset_kmers = parser.copy()
    dataset_kmers.add_argument("distance", type=int, required=False, location="values", 
                              help="maximum distance, default 1000")
    dataset_kmers.add_argument("minimumFrequency", type=int, required=False, location="values", 
                              help="minimum frequency, default 1")
    
    @namespace.doc(description="Get path from k-mer to first splitting k-mer")
    @namespace.expect(dataset_kmers)
    @namespace.doc(params={"uid": "unique identifier dataset","kmer": "k-mer"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,kmer):
        distance = max(1,int(request.args.get("distance",1000)))
        minimumFrequency = max(1,int(request.args.get("minimumFrequency",1)))
        try:
            dsData = _getDataset(uid)
            if dsData:
                if not dsData["collection_location"]==None:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                dsData["collection_location"],dsData["dataset_location"],"kmer.kmc")
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                dsData["collection_location"],dsData["dataset_location"],"kmer.data.h5")
                else:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                dsData["dataset_location"],"kmer.kmc")
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                dsData["dataset_location"],"kmer.data.h5")
                location_kmc = os.path.abspath(location_kmc)
                location_split = os.path.abspath(location_split)
                if not os.path.isfile("{}.kmc_pre".format(location_kmc)):
                    abort(500,"kmc database not found")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                kmc_query_library = haplotyping.service.API.get_kmc_query_library()
                kmc_query_binary_location = haplotyping.service.API.get_kmc_query_binary_location()
                
                newKmer = kmer
                result = {"distance": None, "splittingKmers":[], "pathKmers": [newKmer]}
                for i in range(distance):
                    rightNeighbours = set()
                    leftSplitters = set()
                    for rb in ["A","C","G","T"]:
                        rightNeighbours.add(newKmer[1:]+rb)
                    if i>0:
                        for lb in ["A","C","G","T"]:
                            if not lb==newKmer[0]:
                                leftSplitters.add(lb+newKmer[1:])
                    kmerList = list(rightNeighbours.union(leftSplitters))
                    if kmc_query_library:
                        data = KmerKMC.kmc_library(kmc_query_library,location_kmc,kmerList,0)
                        if not data:
                            abort(500,"no response using kmc query library for "+str(kmerList))
                    elif kmc_query_binary_location:
                        kmc_query_binary_location_query = os.path.join(kmc_query_binary_location,"kmc_query")
                        kmc_query_binary_location_analysis = os.path.join(kmc_query_binary_location,"kmc_analysis")
                        data = KmerKMC.kmc_binary_frequencies(kmc_query_binary_location_query,location_kmc,kmerList,0)
                        if not data:
                            abort(500,"no response using kmc query binary for "+str(kmerList))
                    else:
                        abort(500,"no kmc binary or library configured")
                    #analyse result
                    kmerFound = [k for k in data["kmers"].keys() if data["kmers"][k]>=minimumFrequency]
                    if len([k for k in leftSplitters if k in kmerFound])>0:
                        result["splittingKmers"] = [newKmer]
                        result["distance"] = i
                        result["pathKmers"] = result["pathKmers"][:-1]
                        break               
                    else:
                        rightSplitters = [k for k in rightNeighbours if k in kmerFound or 
                                          haplotyping.General.reverse_complement(k) in kmerFound]
                        if len(rightSplitters)>1:
                            result["splittingKmers"] = list(rightSplitters)
                            result["distance"] = i+1
                            break
                        elif len(rightSplitters)==0:
                            return result
                        else:
                            newKmer = rightSplitters[0] 
                            result["pathKmers"].append(newKmer)  
                if len(result["splittingKmers"])>0:
                    checkedSplittingKmers = []
                    for splittingKmer in result["splittingKmers"]:
                        response = Split.kmer_info(location_split, splittingKmer)
                        if response:
                            checkedSplittingKmers.append(splittingKmer)
                    result["splittingKmers"] = checkedSplittingKmers
                return Response(json.dumps(result), mimetype="application/json")                
            else:
                abort(404, "no dataset with k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>/<kmer1>/path/<kmer2>")
class KmerPath(Resource):
    
    dataset_kmers = parser.copy()
    dataset_kmers.add_argument("distance", type=int, required=False, location="values", 
                              help="maximum distance, default 1000")
    dataset_kmers.add_argument("minimumFrequency", type=int, required=False, location="values", 
                              help="minimum frequency, default 1")
    
    @namespace.doc(description="Get path from k-mer 1 to k-mer 2")
    @namespace.expect(dataset_kmers)
    @namespace.doc(params={"uid": "unique identifier dataset","kmer1": "k-mer","kmer2": "k-mer"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,kmer1,kmer2):
        distance = max(1,int(request.args.get("distance",1000)))
        minimumFrequency = max(1,int(request.args.get("minimumFrequency",1)))
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                   data["collection_location"],data["dataset_location"],"kmer.kmc")
                else:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                   data["dataset_location"],"kmer.kmc")                
                location_kmc = os.path.abspath(location_kmc)
                if not os.path.isfile("{}.kmc_pre".format(location_kmc)):
                    abort(500,"kmc database not found")                
                kmc_query_library = haplotyping.service.API.get_kmc_query_library()
                kmc_query_binary_location = haplotyping.service.API.get_kmc_query_binary_location()
                #construct path
                newKmer = kmer1
                path = kmer1
                for i in range(distance):
                    rightNeighbours = set()
                    for rb in ["A","C","G","T"]:
                        rightNeighbours.add(newKmer[1:]+rb)
                    kmerList = list(rightNeighbours)                    
                    if kmc_query_library:
                        data = KmerKMC.kmc_library(kmc_query_library,location_kmc,kmerList,0)
                        if not data:
                            abort(500,"no response using kmc query library for "+str(kmerList))
                    elif kmc_query_binary_location:
                        kmc_query_binary_location_query = os.path.join(
                            kmc_query_binary_location,"kmc_query")
                        kmc_query_binary_location_analysis = os.path.join(
                            kmc_query_binary_location,"kmc_analysis")
                        data = KmerKMC.kmc_binary_frequencies(
                            kmc_query_binary_location_query,location_kmc,kmerList,0)
                        if not data:
                            abort(500,"no response using kmc query binary for "+str(kmerList))
                    else:
                        abort(500,"no kmc binary or library configured")
                    #analyse result
                    kmerFound = [k for k in data["kmers"].keys() if data["kmers"][k]>=minimumFrequency]
                    rightSplitters = [k for k in rightNeighbours if k in kmerFound or 
                                          haplotyping.General.reverse_complement(k) in kmerFound]
                    if len(rightSplitters)==1:
                        newKmer = rightSplitters[0]
                        path = path + newKmer[-1]
                    elif kmer2 in rightSplitters:
                        for rightSplitter in rightSplitters:
                            if rightSplitter==kmer2:
                                path = path + rightSplitter[-1]                                
                    else:
                        return None
                    if kmer2==path[-len(kmer2):]:
                        return Response(json.dumps(path), mimetype="application/json")
                else:
                    return None
            else:
                abort(404, "no dataset with k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
        
            
@namespace.route("/<uid>")
class KmerMultiple(Resource):            
                
    dataset_kmers = namespace.model("k-mer list to get frequencies", {
        "kmers": fields.List(fields.String, attribute="items", default=[],
                            required=True, description="list of k-mers"),
        "mismatches": fields.Integer(min=0,max=2, default=0, 
                            required=False, description="maximum number of mismatches")
    })
                
    @namespace.doc(description="Get k-mer frequencies from dataset defined by uid")
    @namespace.expect(dataset_kmers)
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):
        mm = min(2,max(0,namespace.payload.get("mismatches",0)))
        kmers = namespace.payload.get("kmers",[])
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.kmc")
                else:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.kmc")
                location_kmc = os.path.abspath(location_kmc)
                if not os.path.isfile("{}.kmc_pre".format(location_kmc)):
                    abort(500,"kmc database not found")                
                kmc_query_library = haplotyping.service.API.get_kmc_query_library()
                kmc_query_binary_location = haplotyping.service.API.get_kmc_query_binary_location()
                if kmc_query_library:
                    response = KmerKMC.kmc_library(kmc_query_library,location_kmc,kmers,mm)
                    if not response:
                        abort(500,"no response using kmc query library for "+str(kmers))
                elif kmc_query_binary_location:
                    kmc_query_binary_location_query = os.path.join(kmc_query_binary_location,"kmc_query")
                    kmc_query_binary_location_analysis = os.path.join(kmc_query_binary_location,"kmc_analysis")
                    response = KmerKMC.kmc_binary_frequencies(kmc_query_binary_location_query,location_kmc,kmers,mm)
                    if not response:
                        abort(500,"no response using kmc query binary for "+str(kmers))
                    response["info"] = KmerKMC.kmc_binary_info(kmc_query_binary_location_analysis,location_kmc)
                else:
                    abort(500,"no kmc binary or library configured")
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
        
            
@namespace.route("/<uid>/sequence")
class KmerSequence(Resource):            
                
    sequence_data = namespace.model("sequence to get k-mer frequencies", {"sequence": 
                                            fields.String(required=False, default="", 
                                            description="sequence"),
                                    "sequences": fields.List(fields.String, attribute="items", 
                                            required=False, default=[],
                                            description="list of sequences"),           
                                    "mismatches": fields.Integer(min=0, max=2, 
                                            default=0, required=False, 
                                            description="maximum number of mismatches")})
    
    @namespace.doc(description="Get k-mer frequencies for a sequence from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @namespace.expect(sequence_data)
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):  
        mm = min(2,max(0,namespace.payload.get("mismatches",0)))
        sequence = namespace.payload.get("sequence","")
        sequences = namespace.payload.get("sequences",[])
        def getKmers(k,sequence,sequences):
            kmers = []
            if k>0:
                for i in range(len(sequence)-(k-1)):
                    kmer = sequence[i:i+k]
                    if not kmer in kmers:
                        kmers.append(kmer)
                for item in sequences:
                    for i in range(len(item)-(k-1)):
                        kmer = item[i:i+k]
                        if not kmer in kmers:
                            kmers.append(kmer)
            return kmers
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                     data["collection_location"],data["dataset_location"],"kmer.kmc")
                else:
                    location_kmc = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                     data["dataset_location"],"kmer.kmc")                    
                location_kmc = os.path.abspath(location_kmc)
                if not os.path.isfile("{}.kmc_pre".format(location_kmc)):
                    abort(500,"kmc database not found")                
                kmc_query_library = haplotyping.service.API.get_kmc_query_library()
                kmc_query_binary_location = haplotyping.service.API.get_kmc_query_binary_location()
                if kmc_query_library:
                    response = KmerKMC.kmc_library(kmc_query_library,location_kmc,[],0)
                    if not response:
                        abort(500,"no response using kmc query library")
                    k = response["info"].get("kmer_length",0)
                    kmers = getKmers(k,sequence,sequences)
                    response = KmerKMC.kmc_library(kmc_query_library,location_kmc,kmers,mm)
                    if not response:
                        abort(500,"no response using kmc query library for "+str(kmers))
                elif kmc_query_binary_location:
                    kmc_query_binary_location_query = os.path.join(
                        kmc_query_binary_location,"kmc_query")
                    kmc_query_binary_location_analysis = os.path.join(
                        kmc_query_binary_location,"kmc_analysis")
                    info = KmerKMC.kmc_binary_info(kmc_query_binary_location_analysis,location_kmc)
                    k = info.get("kmer_length",0)
                    kmers = getKmers(k,sequence,sequences)
                    response = KmerKMC.kmc_binary_frequencies(
                        kmc_query_binary_location_query,location_kmc,kmers,mm)
                    response["info"] = info;
                    if not response:
                        abort(500,"no response using kmc query binary for "+str(kmers))
                else:
                    abort(500,"no kmc binary or library configured")
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))


from flask import Response, abort
from flask_restx import Namespace, Resource
import json, haplotyping

namespace = Namespace("tools", description="Several tools", path="/tools")

@namespace.route("/canonical/<kmer>")
@namespace.param("kmer", "k-mer")
class ToolsCanonicalSingle(Resource):
    @namespace.doc(description="Get canonical form for k-mer")
    def get(self, kmer):
        try:
            response = haplotyping.General.canonical(kmer)
            return Response(json.dumps(response), mimetype="application/json") 
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))

@namespace.route("/reverse-complement/<kmer>")
@namespace.param("kmer", "k-mer")
class ToolsReverseComplement(Resource):
    @namespace.doc(description="Get reverse-complement of k-mer")
    def get(self, kmer):
        try:
            response = haplotyping.General.reverse_complement(kmer)
            return Response(json.dumps(response), mimetype="application/json")
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))



from flask import Response, request, abort
from flask_restx import Namespace, Resource, fields
import json, haplotyping, sqlite3, os

from haplotyping.service.marker import Marker

namespace = Namespace("marker", description="Marker information for a dataset", path="/marker")
parser = namespace.parser()

def getDataset(uid):
    db_connection = haplotyping.service.API.get_db_connection()
    db_connection.row_factory = sqlite3.Row
    cursor = db_connection.cursor()
    cursor.execute("SELECT `dataset`.`uid`, \
                    `dataset`.`type`, \
                    `dataset`.`subtype`, \
                    `dataset`.`version`, \
                    `dataset`.`internal_id`, \
                    `dataset`.`location` AS `dataset_location`, \
                    `collection`.`location` AS `collection_location` \
                    FROM `dataset` \
                    LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                    WHERE `dataset`.`uid` = ? \
                    AND `dataset`.`type` = 'marker' \
                    AND NOT `collection`.`location` IS NULL",(str(uid),))  
    data = cursor.fetchone()
    return data
    
            
@namespace.route("/<uid>/data")
class MarkerData(Resource):
    
    @namespace.doc(description="Get marker data from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    def get(self,uid):
        try:
            data = getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_marker = os.path.join(haplotyping.service.API.get_data_marker_location(),
                                                data["collection_location"],data["dataset_location"])
                else:
                    location_marker = os.path.join(haplotyping.service.API.get_data_marker_location(),
                                                data["dataset_location"])
                marker_id = data["internal_id"]
                (name,markerList,) = Marker.getMarkerData(location_marker,marker_id)
                response=markerList
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with markers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>/mapping")
class MarkerMapping(Resource):
    
    @namespace.doc(description="Get marker mapping for the collection containing the dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    def get(self,uid):
        try:
            data = getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_marker = os.path.join(haplotyping.service.API.get_data_marker_location(),
                                                data["collection_location"],data["dataset_location"])
                else:
                    location_marker = os.path.join(haplotyping.service.API.get_data_marker_location(),
                                                data["dataset_location"])
                markerMapping = Marker.getMarkerMapping(location_marker)
                response=markerMapping
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with markers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            

@namespace.route("/<uid>/info")
class MarkerInfo(Resource):
    
    @namespace.doc(description="Get marker information like reference/alternative "+
                   "and flanking sequences for the collection containing the dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    def get(self,uid):
        try:
            data = getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_marker = os.path.join(haplotyping.service.API.get_data_marker_location(),
                                                data["collection_location"],data["dataset_location"])
                else:
                    location_marker = os.path.join(haplotyping.service.API.get_data_marker_location(),
                                                data["dataset_location"])
                markerInfo = Marker.getMarkerInfo(location_marker)
                response=markerInfo
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with markers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
            


from flask import Response, request, abort
from flask_restx import Namespace, Resource, fields
from flask_caching import Cache
import json, haplotyping, sqlite3, os

from haplotyping.service.split import Split

def _make_cache_key(*args, **kwargs):
    #todo: find better solution
    try:
        cacheKey = "%s_%s_%s_%s" % (args[0].__class__.__name__, str(request.path), str(request.args), str(namespace.payload))
    except:
        cacheKey = "%s_%s_%s" % (args[0].__class__.__name__, str(request.path), str(request.args))
    return cacheKey
   
namespace = Namespace("split", description="Splitting k-mer information for a dataset", path="/split")
cache = Cache()
parser = namespace.parser()

def _getDataset(uid):
    db_connection = haplotyping.service.API.get_db_connection()
    db_connection.row_factory = sqlite3.Row
    cursor = db_connection.cursor()
    cursor.execute("SELECT `dataset`.`uid`, \
                    `dataset`.`type`, \
                    `dataset`.`subtype`, \
                    `dataset`.`version`, \
                    `dataset`.`location` AS `dataset_location`, \
                    `collection`.`location` AS `collection_location` \
                    FROM `dataset` \
                    LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                    WHERE `dataset`.`uid` = ? \
                    AND (`dataset`.`type` = 'kmer' OR `dataset`.`type` = 'split') \
                    AND NOT `collection`.`location` IS NULL",(str(uid),))  
    data = cursor.fetchone()
    return data
            
@namespace.route("/<uid>/info")
class SplitKmerInfo(Resource):
    
    @namespace.doc(description="Get information describing dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.info(location_split)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))

@namespace.route("/<uid>/kmer/distribution")
class SplitKmerDistribution(Resource):
    
    @namespace.doc(description="Get splitting k-mer frequency distribution from dataset defined by uid")
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_split_distribution(location_split)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>/kmer/<kmer>")
class SplitKmerSingle(Resource):
    
    @namespace.doc(description="Get splitting k-mer information from dataset defined by uid for specified k-mer")
    @namespace.doc(params={"uid": "unique identifier dataset","kmer": "splitting k-mer"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,kmer):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_info(location_split, kmer)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>/kmer")
class SplitKmerMultiple(Resource):
    
    dataset_kmers = namespace.model("k-mer list to get splitting k-mer information", {
        "kmers": fields.List(fields.String, attribute="items", required=True, description="list of k-mers")
    })
    
    @namespace.doc(description="Get splitting k-mer information for a list of k-mers from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @namespace.expect(dataset_kmers)
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):
        kmers = namespace.payload.get("kmers",[])
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_list_info(location_split, kmers)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))

@namespace.route("/<uid>/kmer/direct/distribution")
class SplitKmerDirectDistribution(Resource):
    
    @namespace.doc(description="Get splitting k-mer direct connections distance distribution from dataset defined by uid")
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_split_direct_distribution(location_split)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
                
@namespace.route("/<uid>/kmer/direct/<kmer>")
class SplitKmerDirectSingle(Resource):
    
    @namespace.doc(description="Get direct connections splitting k-mer from dataset defined by uid for specified k-mer")
    @namespace.doc(params={"uid": "unique identifier dataset","kmer": "splitting k-mer"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,kmer):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_direct(location_split, kmer)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))    
            
@namespace.route("/<uid>/kmer/direct")
class SplitKmerDirectMultiple(Resource):
    
    dataset_kmers = namespace.model("k-mer list to get direct connections splitting k-mer", {
        "kmers": fields.List(fields.String, attribute="items", required=True, description="list of k-mers")
    })
    
    @namespace.doc(description="Get direct connections splitting k-mer for a list of k-mers from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @namespace.expect(dataset_kmers)
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):
        kmers = namespace.payload.get("kmers",[])
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_list_direct(location_split, kmers)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))  
            
@namespace.route("/<uid>/kmer/read/<kmer>")
class SplitKmerReadSingle(Resource):
    
    @namespace.doc(description="Get read connections splitting k-mer from dataset defined by uid for specified k-mer")
    @namespace.doc(params={"uid": "unique identifier dataset","kmer": "splitting k-mer"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,kmer):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_read(location_split, kmer)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))    
            
@namespace.route("/<uid>/kmer/read")
class SplitKmerReadMultiple(Resource):
    
    dataset_kmers = namespace.model("k-mer list to get read connections splitting k-mer", {
        "kmers": fields.List(fields.String, attribute="items", required=True, description="list of k-mers")
    })
    
    @namespace.doc(description="Get read connections splitting k-mer for a list of k-mers from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @namespace.expect(dataset_kmers)
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):
        kmers = namespace.payload.get("kmers",[])
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_list_read(location_split, kmers)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))  
            
@namespace.route("/<uid>/kmer/paired/<kmer>")
class SplitKmerPairedSingle(Resource):
    
    @namespace.doc(description="Get paired connections splitting k-mer from dataset defined by uid for specified k-mer")
    @namespace.doc(params={"uid": "unique identifier dataset","kmer": "splitting k-mer"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,kmer):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_paired(location_split, kmer)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))    
            
@namespace.route("/<uid>/kmer/paired")
class SplitKmerPairedMultiple(Resource):
    
    dataset_kmers = namespace.model("k-mer list to get paired connections splitting k-mer", {
        "kmers": fields.List(fields.String, attribute="items", required=True, description="list of k-mers")
    })
    
    @namespace.doc(description="Get paired connections splitting k-mer for a list of k-mers from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @namespace.expect(dataset_kmers)
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):
        kmers = namespace.payload.get("kmers",[])
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_list_paired(location_split, kmers)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))  
            
@namespace.route("/<uid>/kmer/sequence")
class SplitKmerSequence(Resource):
    
    sequence_data = namespace.model("sequence to get splitting k-mer information", {"sequence": 
                                            fields.String(required=True, description="sequence")})
    
    @namespace.doc(description="Get splitting k-mer information for a sequence from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @namespace.expect(sequence_data)
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):
        sequence = namespace.payload.get("sequence","")
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_sequence_info(location_split, sequence)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))   

@namespace.route("/<uid>/base/distribution")
class SplitBaseDistribution(Resource):
    
    @namespace.doc(description="Get splitting k-mer base frequency distribution from dataset defined by uid")
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.kmer_base_distribution(location_split)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
            
@namespace.route("/<uid>/base/<base>")
class SplitBaseSingle(Resource):
    
    @namespace.doc(description="Get splitting k-mer base information from dataset defined by uid for specified k-mer base")
    @namespace.doc(params={"uid": "unique identifier dataset","base": "splitting k-mer base"})
    @cache.cached(make_cache_key=_make_cache_key)
    def get(self,uid,base):
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.base_info(location_split, base)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>/base")
class SplitBaseMultiple(Resource):
    
    dataset_bases = namespace.model("k-mer base list to get splitting k-mer base information", {
        "bases": fields.List(fields.String, attribute="items", required=True, description="list of bases")
    })
    
    @namespace.doc(description="Get splitting k-mer base information for a list of k-mer bases from dataset defined by uid")
    @namespace.doc(params={"uid": "unique identifier dataset"})
    @namespace.expect(dataset_bases)
    @cache.cached(make_cache_key=_make_cache_key)
    def post(self,uid):
        bases = namespace.payload.get("bases",[])
        try:
            data = _getDataset(uid)
            if data:
                if not data["collection_location"]==None:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["collection_location"],data["dataset_location"],"kmer.data.h5")
                else:
                    location_split = os.path.join(haplotyping.service.API.get_data_kmer_location(),
                                                data["dataset_location"],"kmer.data.h5")
                if not os.path.isfile("{}".format(location_split)):
                    abort(500,"split database not found")
                response = Split.base_list_info(location_split, bases)
                return Response(json.dumps(response), mimetype="application/json")                
            else:
                abort(404, "no dataset with splitting k-mers for uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))   
            



from flask import Response, request, abort
from flask_restx import Namespace, Resource, abort, fields
import json, haplotyping, sqlite3

namespace = Namespace("dataset", description="All available datasets", path="/dataset")
parser = namespace.parser()

def _make_bool(text):
    return str(text).lower() in ("yes", "true", "t", "1")

def year_description(year_min,year_max):
    if year_min==None and year_max==None:
        return None
    elif year_min==None:
        return "<"+str(year_max+1)
    elif year_max==None:
        return ">"+str(year_min-1)
    elif year_min==year_max:
        return str(year_min)
    else:
        return str(year_min)+"-"+str(year_max)

def adjust_dataset_response(item):
    #variety
    if item["variety_uid"]:
        item["variety"] = {
            "uid": item["variety_uid"], 
            "name": item["variety_name"]
        }
        if item["variety_origin"]:
            item["variety"]["origin"] = {"uid": item["variety_origin"], "country": item["variety_country"]}                
        if item["variety_year_min"] or item["variety_year_max"]:
            item["variety"]["year"] = {"description": year_description(item["variety_year_min"],item["variety_year_max"]), 
                                       "min": item["variety_year_min"], 
                                       "max": item["variety_year_max"]}
    else:
        item["variety"] = None
    del item["variety_uid"]
    del item["variety_name"]
    del item["variety_origin"]
    del item["variety_country"]
    del item["variety_year_min"]
    del item["variety_year_max"]
    #collection
    if item["collection_uid"]:
        item["collection"] = {
            "uid": item["collection_uid"], 
            "name": item["collection_name"]
        }
    else:
        item["collection"] = None
    del item["collection_uid"]
    del item["collection_name"]    
    #finished
    return item


    
@namespace.route("/")
class DatasetList(Resource):

    dataset_list = parser.copy()
    dataset_list.add_argument("start", type=int, required=False, location="values", 
                              help="paging")       
    dataset_list.add_argument("number", type=int, required=False, location="values", 
                              help="paging")
    dataset_list.add_argument("collection", type=str, required=False, location="values", 
                              help="variety has dataset from comma separated list of collection uids")
    dataset_list.add_argument("hasVariety", type=bool, required=False, location="values", 
                              help="linked to a variety")
    dataset_list.add_argument("dataType", type=str, required=False, location="values", choices=["marker","kmer","split"], 
                              help="type of dataset")
    
    dataset_set = namespace.model("uid list to get datasets", {"uids": fields.List(fields.String, attribute="items", 
                          required=True, description="list of uids")})
    
    @namespace.doc(description="Get datasets")    
    @namespace.expect(dataset_list)
    def get(self):
        try:
            start = int(request.args.get("start",0))
            number = int(request.args.get("number",10))        
            name = request.args.get("name",None)
            collection = request.args.get("collection",None)
            hasVariety = request.args.get("hasVariety",None)
            dataType = request.args.get("dataType",None)
            condition_sql = "NOT (`dataset`.`uid` IS NULL OR `dataset`.`type` IS NULL)"
            condition_variables = []
            if not collection==None:
                collection_list = collection.split(",")
                condition_sql = condition_sql + " AND (`collection`.`uid` IN ("+",".join(["?"]*len(collection_list))+"))"
                condition_variables.extend(collection_list)
            if not hasVariety==None:
                if _make_bool(hasVariety):
                    condition_sql = condition_sql + " AND NOT (`dataset`.`variety` IS NULL)"
                else:
                    condition_sql = condition_sql + " AND (`dataset`.`variety` IS NULL)"       
            if not dataType==None:
                if dataType=="kmer":
                    condition_sql = condition_sql + " AND ((`dataset`.`type` IS 'kmer') OR (`dataset`.`type` IS 'split'))"  
                elif dataType=="split":
                    condition_sql = condition_sql + " AND (`dataset`.`type` IS 'split')" 
                elif dataType=="marker":
                    condition_sql = condition_sql + " AND (`dataset`.`type` IS 'marker')"
                else:
                    abort(422, "incorrect dataType condition "+str(dataType))   
            db_connection = haplotyping.service.API.get_db_connection()
            db_connection.row_factory = sqlite3.Row
            cursor = db_connection.cursor()
            cursor.execute("SELECT COUNT(DISTINCT `dataset`.`id`) AS `number` \
                            FROM `dataset` \
                            LEFT JOIN `variety` ON `dataset`.`variety` = `variety`.`uid` \
                            LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` \
                            LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                            WHERE "+condition_sql, tuple(condition_variables))  
            total = cursor.fetchone()[0]
            if start<total:
                cursor.execute("SELECT `dataset`.`uid`, \
                                `dataset`.`accession`, \
                                `dataset`.`type`, \
                                `dataset`.`subtype`, \
                                `dataset`.`version`, \
                                `collection`.`uid` AS `collection_uid`, \
                                `collection`.`name` AS `collection_name`, \
                                `variety`.`uid` AS `variety_uid`, \
                                `variety`.`name` AS `variety_name`, \
                                `variety`.`origin` AS `variety_origin`, \
                                `country`.`name` AS `variety_country`, \
                                `variety`.`year_min` AS `variety_year_min`, \
                                `variety`.`year_max` AS `variety_year_max` \
                                FROM `dataset` \
                                LEFT JOIN `variety` ON `dataset`.`variety` = `variety`.`uid` \
                                LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` \
                                LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                                WHERE "+condition_sql+" \
                                GROUP BY `dataset`.`id` \
                                ORDER BY `dataset`.`uid` \
                                LIMIT ?,?",tuple(condition_variables + [start,number]))  
                resultList = [dict(row) for row in cursor.fetchall()]
            else:
                resultList = []
            for i in range(len(resultList)):
                resultList[i] = adjust_dataset_response(resultList[i])                
            response = {"start": start, "number": number, "total": total, "list": resultList}
            return Response(json.dumps(response), mimetype="application/json") 
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
    @namespace.doc(description="Get multiple datasets by uid")    
    @namespace.expect(dataset_set)
    def post(self):
        uids = namespace.payload.get("uids",[])
        try:
            if len(uids)>0:
                condition_sql = "`dataset`.`uid` IN ("+(",".join(["?"]*len(uids)))+")"
                condition_variables = [] + uids;
                db_connection = haplotyping.service.API.get_db_connection()
                db_connection.row_factory = sqlite3.Row
                cursor = db_connection.cursor()
                cursor.execute("SELECT COUNT(DISTINCT `dataset`.`id`) AS `number` \
                                FROM `dataset` \
                                LEFT JOIN `variety` ON `dataset`.`variety` = `variety`.`uid` \
                                LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` \
                                LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                                WHERE "+condition_sql, tuple(condition_variables))  
                total = cursor.fetchone()[0]
                cursor.execute("SELECT `dataset`.`uid`, \
                                `dataset`.`accession`, \
                                `dataset`.`type`, \
                                `dataset`.`subtype`, \
                                `dataset`.`version`, \
                                `collection`.`uid` AS `collection_uid`, \
                                `collection`.`name` AS `collection_name`, \
                                `variety`.`uid` AS `variety_uid`, \
                                `variety`.`name` AS `variety_name`, \
                                `variety`.`origin` AS `variety_origin`, \
                                `country`.`name` AS `variety_country`, \
                                `variety`.`year_min` AS `variety_year_min`, \
                                `variety`.`year_max` AS `variety_year_max` \
                                FROM `dataset` \
                                LEFT JOIN `variety` ON `dataset`.`variety` = `variety`.`uid` \
                                LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` \
                                LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                                WHERE "+condition_sql+" \
                                GROUP BY `dataset`.`id` \
                                ORDER BY `dataset`.`uid`",tuple(condition_variables))  
                resultList = [dict(row) for row in cursor.fetchall()]
            else:
                total = 0
                resultList = []
            for i in range(len(resultList)):
                resultList[i] = adjust_dataset_response(resultList[i])                
            response = {"total": total, "list": resultList}
            return Response(json.dumps(response), mimetype="application/json") 
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))
            
@namespace.route("/<uid>")
@namespace.doc(params={"uid": "unique identifier dataset"})
class DatasetId(Resource):
    
    @namespace.doc(description="Get dataset by uid")
    def get(self,uid):
        try:
            db_connection = haplotyping.service.API.get_db_connection()
            db_connection.row_factory = sqlite3.Row
            cursor = db_connection.cursor()
            cursor.execute("SELECT `dataset`.`uid`, \
                            `dataset`.`accession`, \
                            `dataset`.`type`, \
                            `dataset`.`subtype`, \
                            `dataset`.`version`, \
                            `collection`.`uid` AS `collection_uid`, \
                            `collection`.`name` AS `collection_name`, \
                            `variety`.`uid` AS `variety_uid`, \
                            `variety`.`name` AS `variety_name`, \
                            `variety`.`origin` AS `variety_origin`, \
                            `country`.`name` AS `variety_country`, \
                            `variety`.`year_min` AS `variety_year_min`, \
                            `variety`.`year_max` AS `variety_year_max` \
                            FROM `dataset` \
                            LEFT JOIN `variety` ON `dataset`.`variety` = `variety`.`uid` \
                            LEFT JOIN `country` ON `variety`.`origin` = `country`.`uid` \
                            LEFT JOIN `collection` ON `dataset`.`collection_id` = `collection`.`id` \
                            WHERE `dataset`.`uid` = ? \
                            GROUP BY `dataset`.`id`",(uid,))  
            data = cursor.fetchone()
            if data:
                response = adjust_dataset_response(dict(data))
                return Response(json.dumps(response), mimetype="application/json")
            else:
                abort(404, "no dataset with uid "+str(uid))
        except Exception as e:
            abort(e.code if hasattr(e,"code") else 500, str(e))


import setuptools
    
with open("haplotyping/_version.py", "r") as fh:
    exec(fh.read())

setuptools.setup(
    name="haplotyping", 
    version=__version__,
    author="Matthijs Brouwer",
    author_email="matthijs.brouwer@wur.nl",
    long_description="k-mer based haplotyping software",
    long_description_content_type="text/plain",
    description="Haplotyping",
    url="https://github.com/haplotyping/haplotyping",
    packages=setuptools.find_packages(exclude=["tests","tests.*"]),
    install_requires=[
      "graphviz >= 0.20.1",
      "pandas >= 2.1.0",
      "numpy >= 1.25.2",
      "networkit >= 10.1",
      "h5py >= 3.9.0",
      "tables >= 3.8.0",
      "pyahocorasick >= 2.0.0",
      "Flask == 2.3.3",
      "flask-restx == 1.2.0",
      "waitress >= 2.1.2",
      "Flask-caching >= 2.0.2",
      "openpyxl >= 3.1.2",
      "frictionless == 5.10.5",
      "psutil >= 5.9.5",
      "metis >= 0.2a5",
      "networkit >= 10.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
    include_package_data=True,
    package_data={"": ["data/schema/*.json"]},
)



import logging,html
import haplotyping.graph.baseGraph as baseGraph
from haplotyping.graph.api import APIGraph
from haplotyping.general import General
import haplotyping
from graphviz import Digraph

class Sections():
    """constructing linear ordered sections from candidates in the De Bruijn graph"""
    
    def __init__(self, graph: baseGraph.Graph):
        """
        construct sections
        """
        assert isinstance(graph, baseGraph.Graph)
        #logger
        self._logger = logging.getLogger(__name__)
        
        self._graph = graph
        self._parts = []
        
        self._datasetFrequencies = None
        self._datasetVarieties = None
        
        #create sections
        for baseConnectedCandidates in self._graph.getConnectedCandidates(True):
            self._createSections(baseConnectedCandidates)
            
            
    def getOrientatedCkmers(self):
        orientatedCkmers = set()
        for i in range(self.partNumber()):
            orientatedCkmers.update(self.part(i).getOrientatedCkmers())
        return orientatedCkmers
        
    def getCandidates(self):
        candidates = []
        for orientatedCkmer in self.getOrientatedCkmers():
            if self._graph._orientatedCkmers[orientatedCkmer].candidate():
                candidates.append(orientatedCkmer)
        return set(candidates)
    
    def processDatasetFrequencies(self):
        if isinstance(self._graph, APIGraph):
            self._datasetFrequencies = self._graph.getDatasetFrequencies()
            self._datasetVarieties = self._graph.getDatasetVarieties()      
            for part in self._parts:
                part._processDatasetFrequencies()                
        else:
            self._logger.warning("unsupported for this graph type")
        
    def _createSections(self, connectedCandidates):
        #compute section start entries (not every connected start is necessarily a section start)
        connected = self._graph.getConnected()
        startEntries = connectedCandidates["start"]
        problematicStartEntries = set()
        for orientatedCkmer in startEntries:
            baseLinkedOrientatedCkmers = set()
            for orientatedBase in self._graph._orientatedCkmers[orientatedCkmer]._orientatedBases.values():
                baseLinkedOrientatedCkmers.update(self._graph._orientatedBases[orientatedBase]._orientatedCkmers)
            for otherOrientatedCkmer in baseLinkedOrientatedCkmers:
                otherIncomingConnected = [k for k in connectedCandidates["connected"] 
                                          if connected[otherOrientatedCkmer][k]]
                if len(otherIncomingConnected)>1:
                    problematicStartEntries.add(orientatedCkmer)
                    break
        if len(problematicStartEntries)>0:
            if len(problematicStartEntries)==len(startEntries):
                self._logger.warning("all {} start entries sections problematic".format(len(startEntries)))
            else:
                startEntries = startEntries.difference(problematicStartEntries)
                self._logger.debug("ignore {} start entries, continue with {} for sections".format(
                    len(problematicStartEntries),len(startEntries)))
        #now try to create sections       
        part = self.SectionsPart(self)   
        partOrientatedCkmers = set()
        sectionItem = self.SectionItem(startEntries, connectedCandidates["start"], connectedCandidates["end"],
                                       connectedCandidates["connected"],self._graph)
        part._append(sectionItem)
        partOrientatedCkmers.update(sectionItem._orientatedCkmers)
        while(len(sectionItem._end["ckmers"].difference(connectedCandidates["end"]))>0):
            newStartCkmers = (sectionItem._end["ckmers"]
                              .intersection(connectedCandidates["connected"])
                              .difference(sectionItem._start["ckmers"]))
            sectionItem = self.SectionItem(newStartCkmers, connectedCandidates["start"], connectedCandidates["end"],
                                       connectedCandidates["connected"],self._graph)
            part._append(sectionItem)
            partOrientatedCkmers.update(sectionItem._orientatedCkmers)
        #compute paths
        part._computePaths()
        self._logger.info("detected {} paths in part with {} sections".format(
            part._pathNumber,len(part._sectionItems)))
        #check consistency
        missingConnected = connectedCandidates["connected"].difference(partOrientatedCkmers)
        if len(missingConnected)>0:
            self._logger.warning("sections not covering complete set of connected candidates, {} missing".format(
                len(missingConnected)))
        #add finalized sections
        self._parts.append(part)
        #try to order logically
        self._parts = sorted(self._parts, key=lambda x: 0 if x._order is None else x._order)
                       
    def partNumber(self):
        return len(self._parts)
    
    def part(self, i=0):
        if i>=0 and i<len(self._parts):
            return self._parts[i]
        else:
            return None
        
    def visualize(self, *args, **kwargs):  
        
        g = Digraph(name="cluster_graph")
        graph_label = "Sectioned Graph"
        if self._graph._name:
            graph_label = "{} {}".format(graph_label,html.escape(self._graph._name))
        graph_label = "{} ({} part{})".format(graph_label,len(self._parts),"s" if len(self._parts)>1 else "")
        graph_label = "<<font point-size=\"12\" color=\"grey\">{}</font>>".format(graph_label)
        g.attr(label=graph_label, labelloc="t", nodesep="0", ranksep="0")        
                
        edges = set()
        nodes = {}

        partConfig = kwargs.copy()
        for i in range(len(self._parts)):
            partConfig["containerGraph"] = g
            partConfig["prefix"] = "graph_part_{}".format(i)
            partConfig["label"] = "Part {}".format(i)
            newEdges, newNodes = self._parts[i].visualize(*args, **partConfig)
            edges.update(newEdges)
            for key,value in newNodes.items():
                if key in nodes:
                    for name,nodeKey in newNodes[key].items():
                        nodes[key][name] = nodeKey
                else:
                    nodes[key] = value.copy()
        return g
                       
    
    class SectionsPart():
        
        def __init__(self, sections):
            """
            construct part
            """
            #logger
            self._logger = logging.getLogger(__name__)
            
            self._sections = sections
            self._graph = sections._graph
            self._sectionItems = []
            self._order : int = None
            self._pathNumber : int = None
                
            self._datasets = set()
            self._varieties = set()
                
        def __repr__(self):
            text = "Part graph"
            if self._graph._name:
                text = "{} {}".format(text,self._graph._name)
            text = "{}; containing {} section{} with {} path{}".format(text,len(self._sectionItems),
                   "s" if len(self._sectionItems)>1 else "", self._pathNumber, "s" if self._pathNumber>1 else "")
            return text
            
        def _append(self, item):
            self._sectionItems.append(item)
            if not item._order is None:
                self._order = (item._order if self._order is None else min(self._order,item._order))
                
        def _computePaths(self):
            self._pathNumber = 0
            numbers = {sectionItemPath.getOrientatedCkmers()[0]: 1 for sectionItemPath in self._sectionItems[0]._paths}
            for sectionItem in self._sectionItems:
                extended = set()
                newNumbers = {sectionItemPath.getOrientatedCkmers()[-1]: 0 for sectionItemPath in sectionItem._paths}
                for sectionItemPath in sectionItem._paths:            
                    orientatedCkmers = sectionItemPath.getOrientatedCkmers()
                    if orientatedCkmers[0] in numbers.keys():
                        extended.add(orientatedCkmers[0])
                    newNumbers[orientatedCkmers[-1]] = (newNumbers[orientatedCkmers[-1]] 
                                                               + numbers.get(orientatedCkmers[0],1))
                for orientatedCkmer in numbers:
                    if not orientatedCkmer in extended:
                        self._pathNumber += numbers[orientatedCkmer]
                numbers = newNumbers
            #add all remaining   
            for orientatedCkmer in numbers:
                if orientatedCkmer in sectionItem._end["ckmers"]:
                    self._pathNumber += numbers[orientatedCkmer]         
            
        def _processDatasetFrequencies(self, ignoreDatasets=[]):
            for sectionItem in self._sectionItems:
                for sectionItemPath in sectionItem._paths:
                    orientatedCkmers = sectionItemPath.getOrientatedCkmers()
                    sectionItemPath._datasets = set(self._sections._datasetFrequencies.index)
                    for orientatedCkmer in orientatedCkmers:
                        ckmerDatasets=self._sections._datasetFrequencies.index[
                            self._sections._datasetFrequencies[orientatedCkmer[0]]>0]
                        sectionItemPath._datasets = sectionItemPath._datasets.intersection(ckmerDatasets)
                    sectionItemPath._varieties = set([self._sections._datasetVarieties[ds]["uid"] 
                                                      for ds in sectionItemPath._datasets 
                                                      if ds in self._sections._datasetVarieties])
                    
        def sectionNumber(self):
            return len(self._sectionItems)
        
        def section(self, i):
            if i>=0 and i<self.sectionNumber():
                return self._sectionItems[i]

        def pathNumber(self):
            return self._pathNumber
            
        def paths(self):
            return self.SectionsPartPathIterator(self)
        
        def getOrientatedCkmers(self):
            orientatedCkmers = set()
            for sectionItem in self._sectionItems:
                orientatedCkmers.update(sectionItem._orientatedCkmers)
            return orientatedCkmers
        
        def getCandidates(self):
            candidates = []
            for orientatedCkmer in self.getOrientatedCkmers():
                if self._graph._orientatedCkmers[orientatedCkmer].candidate():
                    candidates.append(orientatedCkmer)
            return set(candidates)
        
        def visualize(self, *args, **kwargs):  

            def connectArm(arm_key, arm, **kwargs):
                #create and connect arm            
                orientatedCkmer = arm.connection()
                orientatedCkmersArm = arm._orientatedCkmers
                if orientatedCkmer in nodes or len(orientatedCkmersArm.intersection(nodes))>0:
                    if len(orientatedCkmersArm.intersection(nodes))>0:
                        for orientatedCkmerArm in orientatedCkmersArm.intersection(nodes):
                            if arm.armType()=="incoming":
                                if(len(set(self._graph._orientatedCkmers[orientatedCkmerArm]._incoming.keys())
                                               .intersection(nodes))>0):
                                    continue
                                if orientatedCkmerArm in endNodes:
                                    pg.edge(arm_key,endNodes[orientatedCkmerArm],style="dashed", 
                                                       color="grey", rankdir="lr", constraint="true")
                                else:
                                    for ckmerKey in nodes[orientatedCkmerArm].values():
                                        pg.edge(arm_key,ckmerKey,style="dashed", 
                                                       color="grey", rankdir="lr", constraint="true")
                            elif arm.armType()=="outgoing":
                                if(len(set(self._graph._orientatedCkmers[orientatedCkmerArm]._outgoing.keys())
                                               .intersection(nodes))>0):
                                    continue
                                if orientatedCkmerArm in startNodes:
                                    pg.edge(startNodes[orientatedCkmerArm],arm_key,style="dashed", 
                                                       color="grey", rankdir="lr", constraint="true")
                                else:
                                    for ckmerKey in nodes[orientatedCkmerArm].values():
                                        pg.edge(ckmerKey,arm_key,style="dashed", 
                                                       color="grey", rankdir="lr", constraint="true")
                    else:
                        for ckmerKey in nodes[orientatedCkmer].values():
                            if arm.armType()=="incoming":
                                pg.edge(arm_key,ckmerKey,style="dashed", 
                                               color="grey", rankdir="lr", constraint="true")
                            elif arm.armType()=="outgoing":
                                pg.edge(ckmerKey,arm_key,style="dashed", 
                                               color="grey", rankdir="lr", constraint="true")
 
            def getNodeKeys(arm,nodes,startNodes,endNodes):
                keyList = []
                orientatedCkmer = arm.connection()
                orientatedCkmersArm = arm._orientatedCkmers
                if orientatedCkmer in nodes or len(orientatedCkmersArm.intersection(nodes))>0:
                    if len(orientatedCkmersArm.intersection(nodes))>0:
                        for orientatedCkmerArm in orientatedCkmersArm.intersection(nodes):
                            if arm.armType()=="incoming":
                                if(len(set(self._graph._orientatedCkmers[orientatedCkmerArm]._incoming.keys())
                                               .intersection(nodes))>0):
                                    continue
                                if orientatedCkmerArm in endNodes:
                                    keyList.append(endNodes[orientatedCkmerArm])
                                else:
                                    for ckmerKey in nodes[orientatedCkmerArm].values():
                                        keyList.append(ckmerKey)
                            elif arm.armType()=="outgoing":
                                if(len(set(self._graph._orientatedCkmers[orientatedCkmerArm]._outgoing.keys())
                                               .intersection(nodes))>0):
                                    continue
                                if orientatedCkmerArm in startNodes:
                                    keyList.append(startNodes[orientatedCkmerArm])
                                else:
                                    for ckmerKey in nodes[orientatedCkmerArm].values():
                                        keyList.append(ckmerKey)
                    else:
                        for ckmerKey in nodes[orientatedCkmer].values():
                            if arm.armType()=="incoming":
                                keyList.append(ckmerKey)
                            elif arm.armType()=="outgoing":
                                keyList.append(ckmerKey)
                return keyList
            
            initConfig = {
                "showArms": False,
                "showPotentialTransposons": False,
                "baseStylePart": "filled",
                "baseFillColorPart": "beige",
                "basePenWidthPart": 1,
                "baseColorPart": "grey",
                "containerGraph": False,
                "prefix": "graph",
                "label": None
            }
            config = kwargs.copy()
            for key, value in initConfig.items():
                if not key in config.keys():
                    config[key] = value
                    
            #create graph
            if config["containerGraph"]:
                g = config["containerGraph"]
            else:
                g = Digraph()

            edges = set()
            nodes = {}
            startNodes = {}
            endNodes = {}
            
            partGraph=g.subgraph(name="cluster_{}".format(config["prefix"]))      
            
            with partGraph as pg:
                if config["label"]:
                    graph_label = config["label"]
                    graph_label = "{} with {} section{}".format(graph_label, self.sectionNumber(), 
                                                                "s" if self.sectionNumber()>1 else "")
                else:
                    graph_label = "{} section{}".format(self.sectionNumber(),
                                                                "s" if self.sectionNumber()>1 else "")
                    if self._graph._name:
                        graph_label = "{} for Graph {}".format(graph_label,self._graph._name)
                graph_label = ("<" + 
                                  "<font point-size=\"12\" color=\"grey\">{}</font><br/>".format(graph_label) +
                                  "<font point-size=\"10\">{:,} paths</font><br/>".format(self._pathNumber) +
                                  ">")
                graph_style = config["baseStylePart"]
                graph_fillcolor = config["baseFillColorPart"]
                graph_penwidth = config["basePenWidthPart"]
                graph_color = config["baseColorPart"]
                pg.attr(label=graph_label, style=graph_style, fillcolor=graph_fillcolor, 
                        color=graph_color, penwidth=str(graph_penwidth), labelloc="t", nodesep="0", ranksep="0")
                
                #sections
                previousPrefix = ""
                newEdges = set()
                newNodes = {}
                sectionConfig = config.copy()
                sectionConfig["containerGraph"] = pg
                sectionConfig["showPotentialTransposons"] = False
                sectionConfig["showArms"] = False
                sectionConfig["showAllBases"] = False
                sectionConfig["showAllNodes"] = False
                for j in range(self.sectionNumber()):
                    sectionItem = self.section(j)
                    section_prefix = "graph_section_{}".format(j)
                    section_label = "Section {}".format(j)
                    sharedStart = (self._sectionItems[j-1]._orientatedCkmers.intersection(sectionItem._orientatedCkmers) 
                                   if j>0 else set())
                    sharedEnd = (self._sectionItems[j+1]._orientatedCkmers.intersection(sectionItem._orientatedCkmers) 
                                   if j<(self.sectionNumber()-2) else set())
                    sectionConfig["prefix"] = section_prefix
                    sectionConfig["label"] = section_label
                    sectionConfig["hideDeadEndBefore"] = sharedStart
                    sectionConfig["hideDeadEndAfter"] = sharedEnd
                    #create section graph
                    previousEdges = newEdges.copy()
                    previousNodes = newNodes.copy()
                    newEdges, newNodes = self._sectionItems[j].visualize(*args, **sectionConfig)
                    edges.update(newEdges)
                    for key,value in newNodes.items():
                        if key in nodes:
                            for name,nodeKey in newNodes[key].items():
                                nodes[key][name] = nodeKey
                        else:
                            nodes[key] = value.copy()
                    #update start and end
                    for key,value in newNodes.items():
                        if key in sharedStart:
                            for name,nodeKey in newNodes[key].items():
                                startNodes[key] = nodeKey
                        if key in sharedEnd:
                            for name,nodeKey in newNodes[key].items():
                                endNodes[key] = nodeKey
                    #link start/end sections
                    for orientatedCkmer in sharedStart:
                        if orientatedCkmer in previousNodes and orientatedCkmer in newNodes:
                            for ckmerKey1 in previousNodes[orientatedCkmer].values():
                                for ckmerKey2 in newNodes[orientatedCkmer].values():
                                    edge_key = (ckmerKey1, ckmerKey2,)
                                    #single edges
                                    if not edge_key in edges:
                                        edges.add(edge_key)
                                        pg.edge(ckmerKey1, ckmerKey2, style="dashed", color="grey", dir="none", rankdir="tb", constraint="true")
                    previousPrefix = section_prefix                          
                #arms
                if config["showArms"]:
                    arms = self._graph.getArms()
                    processedArms = set()
                    orientatedCkmersSections = self.getOrientatedCkmers()
                    #show potential transposons
                    if config["showPotentialTransposons"]:
                        transposonCandidateArms = self._graph._detectTransposonArmCandidates()
                        for i in range(len(transposonCandidateArms)):
                            arm1 = self._graph.getArm(transposonCandidateArms[i][0])
                            arm2 = self._graph.getArm(transposonCandidateArms[i][1])
                            #only if visible
                            orientatedCkmer1 = arm1.connection()
                            orientatedCkmer2 = arm2.connection()
                            orientatedCkmersArm1 = arm1._orientatedCkmers
                            orientatedCkmersArm2 = arm2._orientatedCkmers
                            if ((orientatedCkmer1 in nodes or len(orientatedCkmersArm1.intersection(nodes))>0) and
                                (orientatedCkmer2 in nodes or len(orientatedCkmersArm2.intersection(nodes))>0)):
                                processedArms.add(arm1.id())
                                processedArms.add(arm2.id())
                                arm_key1, arm_key2 = self._graph._visualizeTransposon(pg,i,arm1,arm2,**config)
                                connectArm(arm_key1,arm1,**config)
                                connectArm(arm_key2,arm2,**config)
                    for j in range(len(arms)):
                        arm = arms[j]
                        #only if no potential transposon
                        if arm.id() in processedArms:
                            continue
                        else:
                            processedArms.add(arm.id())
                        #only if visible
                        armNodeKeys = getNodeKeys(arm,nodes,startNodes,endNodes)
                        if len(armNodeKeys)>0:
                            arm_key = self._graph._visualizeArm(pg,arm,**config)
                            for armNodeKey in armNodeKeys:
                                if arm.armType()=="incoming":
                                    pg.edge(arm_key,armNodeKey,style="dashed", 
                                                               color="grey", rankdir="lr", constraint="true")
                                elif arm.armType()=="outgoing":
                                    pg.edge(armNodeKey,arm_key,style="dashed", 
                                                               color="grey", rankdir="lr", constraint="true")
                #show potential transposons
                elif config["showPotentialTransposons"]:
                    transposonCandidateArms = self._graph._detectTransposonArmCandidates()
                    for i in range(len(transposonCandidateArms)):
                        arm1 = self._graph.getArm(transposonCandidateArms[i][0])
                        arm2 = self._graph.getArm(transposonCandidateArms[i][1])
                        orientatedCkmer1 = arm1.connection()
                        orientatedCkmer2 = arm2.connection()
                        orientatedCkmersArm1 = arm1._orientatedCkmers
                        orientatedCkmersArm2 = arm2._orientatedCkmers
                        if ((orientatedCkmer1 in nodes or len(orientatedCkmersArm1.intersection(nodes))>0) and
                            (orientatedCkmer2 in nodes or len(orientatedCkmersArm2.intersection(nodes))>0)):
                            node_keys1 = getNodeKeys(arm1,nodes,startNodes,endNodes)
                            node_keys2 = getNodeKeys(arm2,nodes,startNodes,endNodes)
                            for armNodeKey1 in node_keys1:
                                for armNodeKey2 in node_keys2:
                                    self._graph._visualizeTransposonConnection(pg, armNodeKey1,armNodeKey2, **config)
            if config["containerGraph"]:
                return (list(edges), nodes)
            else:
                return g
            
        
        class SectionsPartPathIterator():
            
            def __init__(self, sectionsPart):
                self._sectionsPart = sectionsPart
                self._n = 0
                self._path = [-1,[]]
                self._starts = []
                #recompute total
                self._sectionsPart._computePaths()
            
            def __iter__(self):
                self._n = 0
                self._path = [-1,[]]
                self._starts = []
                previousEnds = set()
                for j in range(len(self._sectionsPart._sectionItems)):
                    currentEnds = set()
                    for i in range(len(self._sectionsPart._sectionItems[j]._paths)):
                        orientatedCkmers = self._sectionsPart._sectionItems[j]._paths[i].getOrientatedCkmers()
                        if not orientatedCkmers[0] in previousEnds:
                            self._starts.append((j,i))
                        currentEnds.add(orientatedCkmers[-1])
                    previousEnds = currentEnds
                return self
        
            def __next__(self):
                self._computeNextPath()
                orientatedCkmerList = []
                pathSequence = ""
                distance = 0
                for i in range(len(self._path[1])):
                    if self._path[1][i]>=0:
                        path = self._sectionsPart._sectionItems[i]._paths[self._path[1][i]]
                        if len(orientatedCkmerList)>0:
                            newOrientatedCkmerList = path.getOrientatedCkmers()
                            newPathSequence = path.sequence()
                            assert orientatedCkmerList[-1] == newOrientatedCkmerList[0]
                            assert pathSequence[-self._sectionsPart._graph._k:] == newPathSequence[:self._sectionsPart._graph._k]
                            orientatedCkmerList.extend(newOrientatedCkmerList[1:])
                            pathSequence+=newPathSequence[self._sectionsPart._graph._k:]
                        else:
                            orientatedCkmerList.extend(path.getOrientatedCkmers())
                            pathSequence+=path.sequence()
                        distance+=path.distance()
                return self._sectionsPart.SectionPartPath(orientatedCkmerList, pathSequence, distance, self._sectionsPart._graph)
        
            def _computeNextPath(self):
                findNextPath = True
                nItems = len(self._sectionsPart._sectionItems)
                while findNextPath:
                    findNextPath = False
                    if (len(self._path[1])==0) or (self._path[0]==nItems-1):
                        self._path[0]+=1
                        if self._path[0]>=len(self._starts):
                            raise StopIteration
                        else:
                            #initialise first path
                            startItem = self._starts[self._path[0]]
                            self._path[1]=[-1] * nItems
                            self._path[1][startItem[0]] = startItem[1]
                    else:
                        startItem = self._starts[self._path[0]]
                    #only if not already finished
                    if not (startItem[0]==nItems-1):
                        #find last
                        lastItem = None
                        for i in range(startItem[0],nItems):
                            if self._path[1][i]>=0:
                                lastItem = i
                        #go to next path
                        for i in range(lastItem,startItem[0],-1):
                            self._path[1][i]+=1
                            if self._path[1][i]>=len(self._sectionsPart._sectionItems[i]._paths):
                                self._path[1][i]=-1
                            else:
                                break
                        #all paths checked for startitem
                        if (lastItem>startItem[0]) and (self._path[1][startItem[0]+1]<0):
                            #no other paths for this startitem
                            self._path[1] = []
                            findNextPath = True
                        else:
                            #check
                            previousEnd = self._sectionsPart._sectionItems[startItem[0]]._paths[startItem[1]].getOrientatedCkmers()[-1]
                            for i in range(startItem[0]+1,nItems):
                                if self._path[1][i]<0:
                                    self._path[1][i]=0
                                orientatedCkmers = self._sectionsPart._sectionItems[i]._paths[self._path[1][i]].getOrientatedCkmers()
                                if not orientatedCkmers[0]==previousEnd:
                                    for j in range(i+1,nItems):
                                        self._path[1][j] = -1
                                    findNextPath = True
                                    #check if this is end of path, then ok
                                    currentPathStarts = set([path.getOrientatedCkmers()[0] for 
                                                                path in self._sectionsPart._sectionItems[i]._paths])
                                    if not previousEnd in currentPathStarts:
                                        self._path[1][i] = -1
                                        findNextPath = False
                                    break
                                else:
                                    previousEnd = orientatedCkmers[-1]
                            #reset to next start
                            if max(self._path[1][startItem[0]+1:])<0:
                                self._path[1] = []
                                findNextPath = True
        
                self._n += 1
    
        class SectionPartPath():
                
            def __init__(self, orientatedCkmerList, pathSequence, distance, graph):                
                self._orientatedCkmerList = orientatedCkmerList
                self._pathSequence = pathSequence
                self._distance = distance
                self._graph = graph
                
            def distance(self):
                return self._distance
            
            def sequence(self):
                return self._pathSequence
            
            def kmers(self):
                kmers = []
                for i in range(len(self._pathSequence)-self._graph._k+1):
                    kmers.append(self._pathSequence[i:self._graph._k+i])
                return kmers
            
            def getOrientatedCkmers(self):
                return self._orientatedCkmerList
            
            def getCandidates(self):
                candidates = []
                for orientatedCkmer in self.getOrientatedCkmers():
                    if self._graph._orientatedCkmers[orientatedCkmer].candidate():
                        candidates.append(orientatedCkmer)
                return set(candidates)
            
            def n(self):
                return len(self._orientatedCkmerList)
        
    class SectionItem():
    
        def __init__(self, initialCkmers, startCkmers, endCkmers, connectedCkmers, graph):
            """
            construct section item
            """
            #logger
            self._logger = logging.getLogger(__name__)
            
            self._start = {"ckmers": set(), "bases": set()}
            self._end = {"ckmers": set(), "bases": set()}
            self._orientatedCkmers = set()
            self._orientatedBases = set()
            
            self._connectedCkmers = connectedCkmers
            self._connectedBases = set()
            self._startCkmers = startCkmers
            self._endCkmers = endCkmers
            self._graph = graph
            self._order : int = None
            self._paths : None
            self._pathNumber : int = None
            self._optionalPathNumber : int = None
            self._requiredPathNumber : int = None            
            
            self._checkBackwardList = set()
            self._checkForwardList = {} 
            
            self._datasets = set()
            self._varieties = set()
            
            #compute connected bases
            for connectedCkmer in connectedCkmers:
                self._connectedBases.update(self._graph._orientatedCkmers[connectedCkmer]._orientatedBases.values())

            self._logger.debug("create section with {} starting k-mers".format(len(initialCkmers)))
            for initialCkmer in initialCkmers:
                assert initialCkmer in graph._orientatedCkmers
                self._addOrientatedCkmer(initialCkmer,start=True)
                assert initialCkmer in self._orientatedCkmers
                assert initialCkmer in self._start["ckmers"]

            while len(self._checkBackwardList)>0 or len(self._checkForwardList)>0:
            
                self._checkBackwardList = self._checkBackwardList.difference(self._orientatedCkmers)
                while len(self._checkBackwardList)>0:
                    checkBackwardList = self._checkBackwardList.intersection(self._connectedCkmers)
                    self._checkBackwardList = set()
                    self._logger.debug("check backward {}".format(len(checkBackwardList)))
                    for checkBackwardItem in checkBackwardList:
                        self._addOrientatedCkmer(checkBackwardItem)
                        
                #check if section finished
                allEndingCkmers = set()
                for endingCkmers in self._checkForwardList.values():
                    allEndingCkmers.update(endingCkmers)
                allEndingCkmers = allEndingCkmers.difference(self._start["ckmers"])
                allEndingBases = set()
                for endingCkmer in allEndingCkmers:
                    allEndingBases.update(self._graph._orientatedCkmers[endingCkmer]._orientatedBases.values())
                if len(allEndingBases)==1:
                    for endingCkmer in allEndingCkmers:
                        self._end["ckmers"].add(endingCkmer)
                    self._end["bases"].update(allEndingBases)
                    self._logger.debug("close section with {} k-mers".format(len(self._end["ckmers"])))
                    self._checkForwardList = {}                    
            
                if len(self._checkForwardList)>0:                                
                    checkForwardList = list(self._checkForwardList.keys())
                    #first check for section end
                    forwardBases = set()
                    for checkForwardItem in checkForwardList:
                        assert checkForwardItem in self._graph._orientatedCkmers
                        forwardBases.update(self._graph._orientatedCkmers[checkForwardItem]._orientatedBases.values()) 
                    self._logger.debug("check forward {}".format(len(checkForwardList)))
                    for checkForwardItem in checkForwardList:                        
                        self._addOrientatedCkmer(checkForwardItem)

                        
            #compute paths
            self._computePaths()
            
            #set order
            for orientatedCkmer in self._orientatedCkmers:
                if not self._graph._orientatedCkmers[orientatedCkmer]._order is None:
                    self._order = (self._graph._orientatedCkmers[orientatedCkmer]._order if self._order is None else
                                   min(self._graph._orientatedCkmers[orientatedCkmer]._order,self._order))
                    
                
        def __repr__(self):
            text = "Section object graph"
            if self._graph._name:
                text = "{} {}".format(text,self._graph._name)
            text = "{}; containing {} paths".format(text,self._pathNumber)
            return text
                            
        def _addOrientatedCkmer(self, orientatedCkmer, start=False):
            for orientatedBase in self._graph._orientatedCkmers[orientatedCkmer]._orientatedBases.values():
                assert orientatedBase in self._graph._orientatedBases
                if start:
                    self._start["bases"].add(orientatedBase)
                self._orientatedBases.add(orientatedBase)
                assert orientatedCkmer in self._graph._orientatedBases[orientatedBase]._orientatedCkmers
                for orientatedBaseCkmer in self._graph._orientatedBases[orientatedBase]._orientatedCkmers:
                    assert orientatedBaseCkmer in self._graph._orientatedCkmers
                    self._orientatedCkmers.add(orientatedBaseCkmer)
                    if orientatedBaseCkmer in self._checkForwardList.keys():
                        del self._checkForwardList[orientatedBaseCkmer]
                    if start:
                        self._start["ckmers"].add(orientatedBaseCkmer)
                    if orientatedBaseCkmer in self._endCkmers:
                        self._end["ckmers"].add(orientatedBaseCkmer)
                        self._end["bases"].add(orientatedBase)
                    else:
                        for outgoingCkmer in self._graph._orientatedCkmers[orientatedBaseCkmer]._outgoing:
                            outgoingBases = self._graph._orientatedCkmers[outgoingCkmer]._orientatedBases.values()
                            if len(self._connectedBases.intersection(outgoingBases))>0:
                                if not outgoingCkmer in self._orientatedCkmers:
                                    if outgoingCkmer in self._checkForwardList:
                                        self._checkForwardList[outgoingCkmer].add(orientatedBaseCkmer)
                                    else:
                                        self._checkForwardList[outgoingCkmer] = set([orientatedBaseCkmer])
                    if not orientatedBase in self._start["bases"]:
                        for incomingCkmer in self._graph._orientatedCkmers[orientatedBaseCkmer]._incoming:
                            if incomingCkmer in self._connectedCkmers:
                                if not incomingCkmer in self._orientatedCkmers:
                                    self._checkBackwardList.add(incomingCkmer)
                                
        def _computePaths(self):
            self._paths = []
            #starters are connected, within section and sectionStarter or real starter
            sectionStarters = self._connectedCkmers.intersection(self._orientatedCkmers).intersection(self._start["ckmers"])
            realStarters = self._connectedCkmers.intersection(self._orientatedCkmers).intersection(self._startCkmers)
            partialPaths = [{"distance": 0,
                             "sequence": (orientatedCkmer[0] if orientatedCkmer[1]=="forward" 
                                      else General.reverse_complement(orientatedCkmer[0])),
                             "list": [orientatedCkmer]} 
                            for orientatedCkmer in 
                                        sectionStarters.union(realStarters)]
            while len(partialPaths)>0:
                newPartialPaths = []
                for partialPath in partialPaths:
                    if partialPath["list"][-1] in self._end["ckmers"]:
                            self._paths.append(self.SectionItemPath(partialPath["list"],
                                                               partialPath["sequence"],partialPath["distance"],
                                                               self._graph))
                    else:
                        orientatedCkmerInfo = self._graph._orientatedCkmers[partialPath["list"][-1]]
                        newPaths = 0
                        for orientatedCkmer in orientatedCkmerInfo._outgoing:
                            #ignore cycles
                            if orientatedCkmer in partialPath["list"]:                                
                                pass
                            elif orientatedCkmer in self._orientatedCkmers and orientatedCkmer in self._connectedCkmers:
                                outgoingInfo = orientatedCkmerInfo._outgoing[orientatedCkmer]
                                assert partialPath["sequence"][-self._graph._k:]==outgoingInfo["path"][:self._graph._k]
                                newPartialPath = {"distance": partialPath["distance"] + outgoingInfo["distance"],
                                                  "sequence": partialPath["sequence"] + outgoingInfo["path"][self._graph._k:],
                                                  "list": partialPath["list"] + [orientatedCkmer]}
                                newPartialPaths.append(newPartialPath)
                                newPaths+=1
                        if newPaths==0:
                            self._paths.append(self.SectionItemPath(partialPath["list"],
                                                               partialPath["sequence"],partialPath["distance"],
                                                               self._graph))
                partialPaths = newPartialPaths 
            #update statistics
            self._pathNumber = len(self._paths)
            self._requiredPathNumber = 0
            self._optionalPathNumber = 0
            frequencies = {orientatedCkmer:[] for orientatedCkmer in 
                           self._connectedCkmers.intersection(self._orientatedCkmers)}
            for i in range(len(self._paths)):
                for orientatedCkmer in self._paths[i].getOrientatedCkmers():
                    frequencies[orientatedCkmer].append(i)
            coveredOrientatedCkmers = set()
            for orientatedCkmer in frequencies:
                if len(frequencies[orientatedCkmer])==1:
                    if not self._paths[frequencies[orientatedCkmer][0]].required():
                        self._paths[frequencies[orientatedCkmer][0]]._setRequired()
                        self._requiredPathNumber+=1
                        coveredOrientatedCkmers.update(self._paths[frequencies[orientatedCkmer][0]].getOrientatedCkmers())
            for i in range(len(self._paths)):
                if not self._paths[i].required():
                    if len(set(self._paths[i].getOrientatedCkmers()).difference(coveredOrientatedCkmers))==0:
                        self._paths[i]._setOptional()
                        self._optionalPathNumber+=1
        
        def pathNumber(self):
            return self._pathNumber
        
        def path(self, i: int):
            if i>=0 and i<self._pathNumber:
                return self._paths[i]
            else:
                return None
            
        def getOrientatedCkmers(self):
            return self._orientatedCkmers                        
        
        def getCandidates(self):
            candidates = []
            for orientatedCkmer in self.getOrientatedCkmers():
                if self._graph._orientatedCkmers[orientatedCkmer].candidate():
                    candidates.append(orientatedCkmer)
            return set(candidates)
            
        def visualize(self, *args, **kwargs):  
            
            initConfig = {
                "baseStyleSection": "filled",
                "baseFillColorSection": "grey98",
                "basePenWidthSection": 1,
                "baseColorSection": "grey",
                "containerGraph": False,
                "prefix": "graph_section",
                "label": None
            }
            config = kwargs.copy()
            for key, value in initConfig.items():
                if not key in config.keys():
                    config[key] = value

            #create graph
            if config["containerGraph"]:
                g = config["containerGraph"]
            else:
                g = Digraph()
                
            sectionGraph=g.subgraph(name="cluster_{}".format(config["prefix"]))
            sectionConfig = config.copy()
            with sectionGraph as sg:
                if config["label"]:
                    section_label = config["label"]
                else:
                    section_label = "Section Graph"
                    if self._graph._name:
                        section_label = "{} {}".format(section_label,self._graph._name)
                path_stats = "{} path{}".format(self._pathNumber, "s" if self._pathNumber>1 else "")
                if self._pathNumber==self._requiredPathNumber and self._optionalPathNumber==0:
                    path_stats = "{}, all required".format(path_stats)
                elif self._requiredPathNumber>0 or self._optionalPathNumber>0:
                    path_stats = "{}, {} required and {} optional".format(path_stats, self._requiredPathNumber, 
                                                                          self._optionalPathNumber)
                section_label = ("<" + 
                                 "<font point-size=\"10\">{}</font><br/>".format(section_label) +
                                 "<font point-size=\"8\">{}</font><br/>".format(path_stats) +
                                 ">")                    
                section_style = config["baseStyleSection"]
                section_fillcolor = config["baseFillColorSection"]
                section_penwidth = config["basePenWidthSection"]
                section_color = config["baseColorSection"]
                sg.attr(label=section_label, style=section_style, fillcolor=section_fillcolor, 
                        color=section_color, penwidth=str(section_penwidth),
                        labelloc="t", nodesep="0", ranksep="0")    

                sectionConfig["containerGraph"] = sg
                sectionConfig["restrictedListOfOrientatedCkmers"] = self._orientatedCkmers
                edges,nodes = self._graph.visualize(*args, **sectionConfig)
            if config["containerGraph"]:
                return (list(edges), nodes)
            else:
                return g

        class SectionItemPath():
            
            def __init__(self, orientatedCkmerList, pathSequence, distance, graph):         
                #initialise
                self._orientatedCkmerList = orientatedCkmerList
                self._pathSequence = pathSequence
                self._distance = distance
                self._graph = graph
                #other variables
                self._required = False
                self._optional = False
                self._datasets = set()
                self._varieties = set()

            def distance(self):
                return self._distance
            
            def sequence(self):
                return self._pathSequence

            def kmers(self):
                kmers = []
                for i in range(len(self._pathSequence)-self._graph._k+1):
                    kmers.append(self._pathSequence[i:self._graph._k+i])
                return kmers
            
            def getOrientatedCkmers(self):
                return self._orientatedCkmerList
            
            def getCandidates(self):
                candidates = []
                for orientatedCkmer in self.getOrientatedCkmers():
                    if self._graph._orientatedCkmers[orientatedCkmer].candidate():
                        candidates.append(orientatedCkmer)
                return set(candidates)
            
            def n(self):
                return len(self._orientatedCkmerList)
                
            def _setRequired(self):
                self._required = True
                self._optional = False

            def required(self):
                return self._required
            
            def _setOptional(self):
                if not self._required:
                    self._optional = True                
                
            def optional(self):
                return self._optional
                
            
                
                
                
                      
        
            
        
        


import logging, h5py, tables, gzip, time
import os, sys, math, signal, psutil
import re, haplotyping
import numpy as np
import haplotyping.index.storage
import haplotyping.index.splits
import haplotyping.index.database
import multiprocessing as mp
from threading import Event
from queue import Empty

class Connections:
    
    """
    Internal use, parse read files and store results in database
    """
    
    stepSizeStorage = 1000000
    
    def __init__(self, unpairedReadFiles, pairedReadFiles, h5file, filenameBase, 
                 indexType=None, debug=False, keepTemporaryFiles=False):
        
        """
        Internal use only: initialize
        """
        
        #logger
        self._logger = logging.getLogger(__name__)
        if len(unpairedReadFiles)>0:
            self._logger.info("parse "+str(len(unpairedReadFiles))+" unpaired readfile(s)")
        if len(pairedReadFiles)>0:
            self._logger.info("parse "+str(2*len(pairedReadFiles))+" paired readfile(s)")                
                    
        self.debug = debug
        self.keepTemporaryFiles = keepTemporaryFiles
        self.indexType = indexType
        self.filenameBase = filenameBase
        
        #set variables
        self.k = h5file["/config"].attrs["k"]
        self.automatonKmerSize = h5file["/config"].attrs["automatonKmerSize"]
        self.maximumFrequency = h5file["/config"].attrs["maximumCanonicalSplitFrequency"]
        self.minimumFrequency = h5file["/config"].attrs["minimumCanonicalSplitFrequency"]
        self.maximumMemory = h5file["/config"].attrs["maximumMemory"]
        self.maximumProcesses = h5file["/config"].attrs["maximumProcesses"]
        self.numberOfKmers = h5file["/split/ckmer"].shape[0]
        self.totalNumberOfKmers = h5file["/config"].attrs["numberKmers"]
        self.h5file = h5file
        self.unpairedReadFiles = unpairedReadFiles
        self.pairedReadFiles = pairedReadFiles
        self.temporaryMergedRelationsFile = None
        
        #statistics
        self.readLengthMinimum=None
        self.readLengthMaximum=None
        self.readPairedTotal=0
        self.readUnpairedTotal=0
        self.readTotal=0
        self.totalReadLength=0
        self.processReadsTime=0
        
        #estimated size
        self.estimatedMaximumReadLength = 0
        
        #check existence group
        if not "/relations" in h5file:
            h5file.create_group("/relations")
        
        #create relations datasets
        if "/relations/direct" in h5file:
            self._logger.warning("direct relation dataset already exists in hdf5 storage")
        elif "/relations/cycle" in h5file:
            self._logger.warning("cycle relation dataset already exists in hdf5 storage")
        elif "/relations/reversal" in h5file:
            self._logger.warning("reversal relation dataset already exists in hdf5 storage")
        else:
            #theoretical maximum is 2 * #letters, however read-errors will introduce additional connections
            #partly they will be pre-filtered, but a buffer seems sensible (?)
            self.numberDirectArray = 2*len(haplotyping.index.Database.letters)
            h5file["/config/"].attrs["numberDirectArray"] = self.numberDirectArray
                                    
            #process
            try:                                                
                #create automaton and index
                automatonKmerSize = (math.ceil((self.k+1)/2) 
                                     if self.automatonKmerSize==0 else min(self.k,self.automatonKmerSize))  
                self.automatonKmerSize = automatonKmerSize
                (automatonMemory,indexFile, automatonFile) = haplotyping.index.splits.Splits.createAutomatonWithIndex(
                    self.h5file, filenameBase, automatonKmerSize)
                #process
                pytablesFile = filenameBase+"_tmp_connections_merge.h5"
                if os.path.exists(pytablesFile):
                    os.remove(pytablesFile)
                self._logger.debug("store temporary in "+pytablesFile)    
                #create datasets
                with tables.open_file(pytablesFile, mode="w", title="Temporary storage") as pytablesStorage:
                    self._processReadFiles(indexFile, automatonFile, automatonMemory, pytablesStorage)
                    self._storeDirect(pytablesStorage)
                    self.h5file.flush()
                    if not self.indexType==haplotyping.index.database.Database.ONLYDIRECTCONNECTIONS:
                        self._processReads(pytablesStorage)
                        self._storeReads(pytablesStorage)
                        self.h5file.flush()     
            #except Exception as e:
            #   self._logger.error("problem occurred while processing reads: "+str(e))
            finally:
                try:
                    if not self.keepTemporaryFiles:
                        os.remove(pytablesFile)
                        haplotyping.index.splits.Splits.deleteAutomatonWithIndex(filenameBase, automatonKmerSize)
                except:
                    self._logger.error("problem removing files")

#--------------
# Handle Queue
#--------------    
                    
    def _collect_queue(queue_entry):
        time.sleep(0.1)
        while True:
            try:
                item = queue_entry.get(block=True, timeout=1)
            except Empty:
                break

    def _close_queue(queue_entry):
        time.sleep(0.1)
        queue_entry.close()
        queue_entry.join_thread()

    def _collect_and_close_queue(queue_entry):
        entries = []
        queue_entry.put(None)
        time.sleep(0.1)
        while True:
            try:
                item = queue_entry.get(block=True, timeout=1)
                if item==None:
                    break
                else:
                    entries.append(item)
            except Empty:
                break 
        #now also close this queue
        Connections._close_queue(queue_entry)
        return entries
    
#---------------
# Handle Memory
#---------------

    def _processMemory():
        process = psutil.Process(os.getpid())
        memory = process.memory_info().rss
        for child in process.children(recursive=True):
            memory += child.memory_info().rss
        return memory
    
#-----------------------------------
# Main functions Direct Connections
#-----------------------------------

    def _processReadFiles(self, indexFile, automatonFile, automatonMemory, pytablesStorage):
          
        def estimateIndexMemory(nWorkersAutomaton,workerAutomatonMemory,nWorkersMatches,
                           workerMatchesMemory,nWorkersIndex,workerIndexMemory):
            return ((nWorkersAutomaton*workerAutomatonMemory) + 
                    (nWorkersMatches*workerMatchesMemory) + 
                    (nWorkersIndex*workerIndexMemory))
                
        #get method
        self._logger.debug("using method '{}' for multiprocessing".format(mp.get_start_method()))
        process = psutil.Process(os.getpid())
        self._logger.debug("initially used memory {} MB".format(math.ceil(Connections._processMemory()/1048576)))
        
        self._logger.debug("memory info: {}".format(psutil.Process(os.getpid()).memory_info()))
        
        #compute size shared memory k-mer index (to confirm results from automaton)
        shm_index_size = os.path.getsize(indexFile)
        self._logger.debug("size shared memory {} MB k-mer index".format(math.ceil(shm_index_size/1048576)))
                    
        #compute size shared memory k-mer type, number and bases
        shm_kmer_link = np.dtype(haplotyping.index.Database.getUint(self.numberOfKmers)).type
        shm_kmer_number = np.dtype(haplotyping.index.Database.getUint(self.maximumFrequency)).type
        shm_kmer_size = self.numberOfKmers*(1+shm_kmer_number(0).nbytes+(2*shm_kmer_link(0).nbytes))
        self._logger.debug("size shared memory {} MB k-mer properties".format(math.ceil(shm_kmer_size/1048576)))
        
        #shutdown
        shutdown_event = mp.Event()
        
        qsize = 10000
        queue_start = mp.Queue(qsize)
        queue_automaton = mp.Queue(qsize)
        queue_index = mp.Queue(qsize)
        queue_matches = mp.Queue(qsize)
        queue_finished = mp.Queue()
        queue_storageDirect = mp.Queue()
        queue_storageReads = mp.Queue()
        
        #estimate worker automaton memory
        workerAutomatonMemory = automatonMemory
        #estimate worker matches memory
        workerMatchesDtypeEntry = haplotyping.index.storage.Storage.worker_matches_dtype(
            self.numberOfKmers,self.maximumFrequency,self.estimatedMaximumReadLength,self.numberDirectArray)
        workerMatchesMemory = self.numberOfKmers * np.dtype(workerMatchesDtypeEntry).itemsize
        #estimate worker index memory (shared)
        workerIndexMemory = 0
        workerSharedMemory = (shm_kmer_size+shm_index_size)
        
        self._logger.debug("estimated shared memory: {} MB".format(math.ceil(workerSharedMemory/1048576)))
        self._logger.debug("estimated memory automaton worker: {} MB".format(math.ceil(workerAutomatonMemory/1048576)))
        self._logger.debug("estimated memory index worker: {} MB".format(math.ceil(workerIndexMemory/1048576)))
        self._logger.debug("estimated memory matches worker: {} MB".format(math.ceil(workerMatchesMemory/1048576)))
        
        #worker requirements
        nWorkers = max(3,mp.cpu_count()-1) if self.maximumProcesses==0 else self.maximumProcesses - 1
        
        #memory requirements
        if self.maximumMemory>0:
            maximumMemory = min(psutil.virtual_memory().available + process.memory_info().rss, self.maximumMemory)
        else:
            maximumMemory = round(0.95*psutil.virtual_memory().available) + process.memory_info().rss    
            
            
        #compute maximum number of automaton workers (high memory usage)
        #assume that ideal ratio workers is 1:2:4 (to be verified/computed?)
        nWorkersAutomaton = max(1,min(math.floor(nWorkers/3),
                                      math.floor((maximumMemory-workerSharedMemory)/
                                             estimateIndexMemory(1,workerAutomatonMemory,
                                                                 2,workerMatchesMemory,
                                                                 4,workerIndexMemory))))
        #auto distribute other workers within limits
        nWorkersMatches = math.floor((nWorkers - nWorkersAutomaton)/2)
        nWorkersIndex = nWorkers - nWorkersAutomaton - nWorkersMatches        
        #increment if processes available
        nWorkersLeft = nWorkers - nWorkersAutomaton - nWorkersMatches - nWorkersIndex
        while(nWorkersLeft>0):
            if(nWorkersLeft>0):
                nWorkersMatches+=1
                nWorkersLeft-=1
            if(nWorkersLeft>0):
                nWorkersIndex+=1
                nWorkersLeft-=1                                         
        #reduce to fit memory requirements   
        while (estimatedMemory := estimateIndexMemory(nWorkersAutomaton,workerAutomatonMemory,
                             nWorkersMatches,workerMatchesMemory,
                             nWorkersIndex,workerIndexMemory)) + workerSharedMemory > maximumMemory:
            if (nWorkersAutomaton==1) and (nWorkersMatches==1) and (nWorkersIndex==1):
                raise Exception("not enough memory available, required: {} MB".format(round(estimatedMemory/1048576)))
            elif (nWorkersAutomaton==1) and (nWorkersMatches==1):
                nWorkersIndex -= 1
            elif (nWorkersMatches>nWorkersAutomaton):
                nWorkersMatches -= 1
                nWorkersIndex = nWorkersIndex+1
            else:
                nWorkersAutomaton-=1
                nWorkersIndex = nWorkersIndex+1
        #don't oversize the index workers, maximum two times matches workers
        nWorkersIndex = min(nWorkersIndex,2*nWorkersMatches)
        #final calculation memory estimation
        estimatedMemory = estimateIndexMemory(nWorkersAutomaton,workerAutomatonMemory,
                             nWorkersMatches,workerMatchesMemory,nWorkersIndex,workerIndexMemory) + workerSharedMemory
        
        self._logger.debug("start {} processes to parse reads with reduced automaton".format(nWorkersAutomaton))
        self._logger.debug("start {} processes to check index".format(nWorkersIndex))
        self._logger.debug("start {} processes to process matches".format(nWorkersMatches))
        self._logger.debug("estimated total memory usage: {} MB".format(math.ceil(estimatedMemory/1048576)))
                
        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        pool_automaton = mp.Pool(nWorkersAutomaton, haplotyping.index.storage.Storage.workerAutomaton, 
                             (shutdown_event,queue_start,queue_automaton,queue_index,queue_finished,
                              self.k,self.automatonKmerSize,automatonFile,))
        #first start automatons, one at a time, because of memory peak
        startedAutomaton = 0
        queue_start.put("automaton")
        while not (startedAutomaton==nWorkersAutomaton):
            try:
                item = queue_finished.get(block=True, timeout=1)
                if item=="automaton:started":
                    startedAutomaton+=1
                    self._logger.debug("{} of {} automatons started".format(startedAutomaton,nWorkersAutomaton))
                    if startedAutomaton<nWorkersAutomaton:
                        queue_start.put("automaton")
                else:
                    self._logger.error("unexpected start value in finished queue: {}".format(item))
            except:
                pass
            time.sleep(1)
            
        #create shared memory k-mer index (to confirm results from automaton)
        shm_index = mp.shared_memory.SharedMemory(create=True, size=shm_index_size)
        with open(indexFile, "r") as f:
            shm_index.buf[0:shm_index_size] = bytes(f.read(),"utf-8")
        self._logger.debug("created shared memory {} MB k-mer index".format(math.ceil(shm_index_size/1048576)))
        
        #create shared memory k-mer type, number and bases
        shm_kmer = mp.shared_memory.SharedMemory(create=True, size=shm_kmer_size)
        kmer_properties = np.ndarray((self.numberOfKmers,), dtype=[("type","S1"),("number",shm_kmer_number),
                           ("left",shm_kmer_link),("right",shm_kmer_link)], buffer=shm_kmer.buf)
        ckmerLink = 0
        for i in range(0,self.numberOfKmers,Connections.stepSizeStorage):
            ckmers = self.h5file["/split/ckmer"][i:min(self.numberOfKmers,i+Connections.stepSizeStorage)]
            stepData = [(row[1],row[2],row[3][0],row[3][1],) for row in ckmers]
            kmer_properties[ckmerLink:ckmerLink+len(stepData)] = stepData
            ckmerLink+=len(stepData)
            del stepData
        self._logger.debug("created shared memory {} MB k-mer properties".format(math.ceil(shm_kmer_size/1048576)))

        #now start other workers
        pool_index = mp.Pool(nWorkersIndex, haplotyping.index.storage.Storage.workerIndex, 
                             (shutdown_event,queue_index,queue_matches,queue_storageReads,queue_finished,
                              self.filenameBase,self.numberOfKmers,self.k,
                              self.indexType,shm_index.name))
        pool_matches = mp.Pool(nWorkersMatches, haplotyping.index.storage.Storage.workerMatches, 
                               (shutdown_event,queue_matches,queue_storageDirect,queue_finished,
                                self.filenameBase,self.numberOfKmers,self.maximumFrequency,
                                self.estimatedMaximumReadLength,self.numberDirectArray,
                                self.indexType,shm_kmer.name))
        signal.signal(signal.SIGINT, original_sigint_handler)

        try:
            #process and register unpaired read files
            if not "unpairedReads" in self.h5file["/config/"].keys():
                dtypeList = [("file","S255"),("readLength","uint64"),
                     ("readNumber","uint64"),("totalReadLength","uint64"),("processTime","uint32")]
            
                ds = self.h5file["/config/"].create_dataset("unpairedReads",(len(self.unpairedReadFiles),),
                                                  dtype=np.dtype(dtypeList),chunks=None, 
                                                  compression="gzip", compression_opts=9)
                for i in range(len(self.unpairedReadFiles)):
                    self._logger.debug("process {}".format(os.path.basename(self.unpairedReadFiles[i])))
                    (readLength,readNumber,totalReadLength,processTime) = self._processReadFile(
                                                                   self.unpairedReadFiles[i], 
                                                                   queue_automaton, queue_index, queue_matches)
                    ds[i] = (self.unpairedReadFiles[i],
                             readLength,readNumber,totalReadLength,int(processTime))
            else:
                self._logger.error("unpairedReads already (partly) processed")

            #process and register paired read files
            if not "pairedReads" in self.h5file["/config/"].keys():
                dtypeList = [("file0","S255"),("file1","S255"),("readLength","uint64"),
                         ("readNumber","uint64"),("totalReadLength","uint64"),("processTime","uint32")]
                ds = self.h5file["/config/"].create_dataset("pairedReads",(len(self.pairedReadFiles),),
                                                  dtype=np.dtype(dtypeList),chunks=None, 
                                                  compression="gzip", compression_opts=9)
                for i in range(len(self.pairedReadFiles)):
                    self._logger.debug("process {} and {}".format(os.path.basename(self.pairedReadFiles[i][0]),
                                                           os.path.basename(self.pairedReadFiles[i][1])))
                    (readLength,readNumber,totalReadLength,processTime) = self._processPairedReadFiles(
                                                                     self.pairedReadFiles[i][0],
                                                                     self.pairedReadFiles[i][1], 
                                                                     queue_automaton, queue_index, queue_matches)
                    ds[i] = (self.pairedReadFiles[i][0],self.pairedReadFiles[i][1],
                             readLength,readNumber,totalReadLength,int(processTime))
            else:
                self._logger.error("pairedReads already (partly) processed")                    
            
            #now wait until queues are empty
            while not (queue_automaton.empty() and queue_index.empty() and queue_matches.empty()):
                time.sleep(1)
                
            #then trigger stopping by sending enough Nones
            for i in range(pool_automaton._processes):
                queue_automaton.put(None)
            for i in range(pool_index._processes):
                queue_index.put(None)
            for i in range(pool_matches._processes):
                queue_matches.put(None)
                
            #now wait until everyone is finished
            finishedAutomaton=0
            finishedIndex=0
            finishedMatches=0
            totalCanonicalSplitFrequencies=0
            while not (finishedAutomaton==nWorkersAutomaton and finishedIndex==nWorkersIndex
                       and finishedMatches==nWorkersMatches):
                try:
                    item = queue_finished.get(block=True, timeout=1)
                    if item.startswith("automaton:ended"):
                        finishedAutomaton+=1
                    elif item.startswith("index:ended"):
                        finishedIndex+=1
                        totalCanonicalSplitFrequencies+=int(item.split(":")[3])
                    elif item.startswith("matches:ended"):
                        finishedMatches+=1
                    else:
                        self._logger.error("unexpected value in finished queue: {}".format(item))
                except:
                    pass
                time.sleep(1)
            #store total found canonical k-mers (doesn't fully match results from kmc because of overlapping sequences)
            self.h5file["/config/"].attrs["totalCanonicalSplitFrequencies"] = totalCanonicalSplitFrequencies
                
        except KeyboardInterrupt:
            self._logger.debug("caught KeyboardInterrupt")
            #collect from queue
            Connections._collect_queue(queue_automaton)
            Connections._collect_queue(queue_index)
            Connections._collect_queue(queue_matches)
            #shutdown directly
            shutdown_event.set()
            #release memory
            shm_index.close()
            shm_kmer.close()
            shm_index.unlink()
            shm_kmer.unlink()
            sys.exit()
        finally:
            #terminate pools
            pool_automaton.terminate()
            pool_index.terminate()
            pool_matches.terminate()
            #close queus workers
            Connections._close_queue(queue_automaton)
            Connections._close_queue(queue_index)
            Connections._close_queue(queue_matches)
            #shutdown
            shutdown_event.set()
            #join pools
            pool_automaton.join()
            pool_index.join()
            pool_matches.join()
            #release memory
            shm_index.close()
            try:
                shm_index.unlink()
            except Exception as e:
                self._logger.debug("problem unlinking shared memory ({})".format(e))
            #collect created files 
            storageDirectFiles = Connections._collect_and_close_queue(queue_storageDirect)
            self.storageReadFiles = Connections._collect_and_close_queue(queue_storageReads)            
            
        #now all files are created, so merging can start
        self._logger.debug("merge {} files with direct connections".format(len(storageDirectFiles)))
            
        #reset shutdown event
        shutdown_event.clear()
        
        #assume memory is no problem
        nWorkersMerges = max(1,mp.cpu_count()-1) if self.maximumProcesses==0 else self.maximumProcesses - 1
        self._logger.debug("start {} processes to merge stored data".format(nWorkersMerges))
        
        queue_ranges = mp.Queue(nWorkersMerges)
        queue_merges = mp.Queue(nWorkersMerges)
        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        pool_merges = mp.Pool(nWorkersMerges, haplotyping.index.storage.Storage.workerMergeDirect, 
                               (shutdown_event,queue_ranges,queue_merges,
                                storageDirectFiles,
                                self.filenameBase,self.numberOfKmers,
                                self.maximumFrequency,self.minimumFrequency,
                                shm_kmer.name))
        signal.signal(signal.SIGINT, original_sigint_handler)
        
        try:
            #fill queue
            mergeNumber = math.ceil(self.numberOfKmers/nWorkersMerges)
            for mergeStart in range(0,self.numberOfKmers,mergeNumber):
                queue_ranges.put((mergeStart,mergeNumber,))

            #then trigger stopping by sending enough Nones
            for i in range(pool_merges._processes):
                queue_ranges.put(None)
                
            #now wait    
            while not (queue_ranges.empty()):
                time.sleep(1)
                
            #clean
            if not self.keepTemporaryFiles:
                for item in storageDirectFiles:
                    os.remove(item)
                                
            #collect created merges
            mergeFiles = []
            while True:
                try:
                    item = queue_merges.get(block=True, timeout=1)
                    if item==None:
                        break
                    elif isinstance(item,str):
                        mergeFiles.append(item)
                except Empty:
                    break 
            #now also close this queue
            Connections._close_queue(queue_merges)
            
            self._logger.debug("combine {} merged files".format(len(mergeFiles)))
            
            #combine
            haplotyping.index.storage.Storage.combineDirectMerges(
                mergeFiles, pytablesStorage, self.numberOfKmers, self.maximumFrequency)                        
            
            #clean
            if not self.keepTemporaryFiles:
                for item in mergeFiles:
                    os.remove(item)
                
        except KeyboardInterrupt:
            self._logger.debug("caught KeyboardInterrupt")
            shutdown_event.set()
            #release memory
            shm_kmer.close()
            shm_kmer.unlink()
            sys.exit()
        finally:
            #terminate pool
            pool_merges.terminate()
            #close queus workers
            Connections._close_queue(queue_ranges)
            Connections._close_queue(queue_merges)
            #shutdown
            shutdown_event.set()
            #join pool
            pool_merges.join()
            #release memory
            shm_kmer.close()
            try:
                shm_kmer.unlink()
            except Exception as e:
                self._logger.debug("problem unlinking shared memory ({})".format(e))
        self._logger.debug("created merged file")
        
        self._logger.debug("process {} files with read information".format(len(self.storageReadFiles)))        
            
    def _processReadFile(self, filename: str, queue_automaton, queue_index, queue_matches):
        startTime = time.time()
        open_fn = gzip.open if filename.endswith(".gz") else open
        with open_fn(filename, "rt") as f:
            readLengthMinimum=None
            readLengthMaximum=None
            readNumber=0
            totalReadLength=0
            while True:               
                identifier = f.readline().rstrip()
                sequence = f.readline().rstrip()  
                plusline = f.readline().rstrip()
                quality = f.readline().rstrip()
                if sequence:
                    if not (identifier.startswith("@") and plusline.startswith("+") and len(sequence)==len(quality)):
                        self._logger.error("invalid fastq-file {}, line {}".format(filename,4*readNumber))
                        break
                    else:
                        readLengthMinimum=(len(sequence) if readLengthMinimum==None 
                                           else min(readLengthMinimum,len(sequence)))
                        readLengthMaximum=(len(sequence) if readLengthMaximum==None 
                                           else max(readLengthMaximum,len(sequence)))
                        readNumber+=1
                        totalReadLength+=len(sequence)
                        queue_automaton.put(sequence)
                        if readNumber%1000000==0:
                            self._logger.debug("- processed {} reads".format(readNumber)) 
                else:
                    break
            endTime = time.time()
            if readNumber>0:
                if readLengthMinimum>0:
                    self.readLengthMinimum=(readLengthMinimum if self.readLengthMinimum==None 
                                            else min(self.readLengthMinimum,readLengthMinimum))
                self.readLengthMaximum=(readLengthMaximum if self.readLengthMaximum==None 
                                        else max(self.readLengthMaximum,readLengthMaximum))
                self.readUnpairedTotal+=readNumber
                self.readTotal+=readNumber
                self.totalReadLength+=totalReadLength
            self.processReadsTime+=endTime-startTime
        self._logger.info("processed {} reads".format(readNumber)) 
        return (readLengthMaximum,readNumber,totalReadLength,endTime-startTime)

    def _processPairedReadFiles(self, filename0: str, filename1: str, queue_automaton, queue_index, queue_matches):
        startTime = time.time()
        open_fn0 = gzip.open if filename0.endswith(".gz") else open
        open_fn1 = gzip.open if filename1.endswith(".gz") else open
        with open_fn0(filename0, "rt") as f0, open_fn1(filename1, "rt") as f1:
            readLengthMinimum=None
            readLengthMaximum=None
            readNumber=0
            totalReadLength=0
            while True:
                #first of pair
                identifier0 = f0.readline().rstrip() 
                sequence0 = f0.readline().rstrip()  
                plusline0 = f0.readline().rstrip() 
                quality0 = f0.readline().rstrip() 
                #second of pair
                identifier1 = f1.readline().rstrip() 
                sequence1 = haplotyping.General.reverse_complement(f1.readline().rstrip())
                plusline1 = f1.readline().rstrip() 
                quality1 = f1.readline().rstrip() 
                #process
                if sequence0 and sequence1:
                    if not (identifier0.startswith("@") and plusline0.startswith("+") and len(sequence0)==len(quality0)):
                        self._logger.error("invalid fastq-file {}, line {}".format(filename0,2*readNumber))
                        break
                    elif not (identifier1.startswith("@") and plusline1.startswith("+") and len(sequence1)==len(quality1)):
                        self._logger.error("invalid fastq-file {}, line {}".format(filename1,2*readNumber))
                        break
                    else:
                        readLengthMinimum=(min(len(sequence0),len(sequence1)) if readLengthMinimum==None 
                                               else min(readLengthMinimum,len(sequence0),len(sequence1)))
                        readLengthMaximum=(max(len(sequence0),len(sequence1)) if readLengthMaximum==None 
                                           else max(readLengthMaximum,len(sequence0),len(sequence1)))
                        readNumber+=2  
                        totalReadLength+=(len(sequence0)+len(sequence1))
                        if sequence1[0:self.k] in sequence0:
                            pos = sequence0.find(sequence1[0:self.k])
                            rpos = sequence0.rfind(sequence1[0:self.k])
                            if pos==rpos:
                                match = sequence0[pos:]
                                if sequence1[0:len(match)]==match:
                                    #process as single read because of minimal glue match of size k
                                    queue_automaton.put(sequence0[0:pos]+sequence1) 
                                else:
                                    queue_automaton.put((sequence0,sequence1,))                                
                            else:
                                queue_automaton.put((sequence0,sequence1,))                            
                        else:
                            queue_automaton.put((sequence0,sequence1,))                        
                        if readNumber%1000000==0:
                            self._logger.debug("- processed {} paired reads".format(readNumber))                     
                else:
                    break
            endTime = time.time()
            if readNumber>0:
                if readLengthMinimum>0:
                    self.readLengthMinimum=(readLengthMinimum if self.readLengthMinimum==None 
                                            else min(self.readLengthMinimum,readLengthMinimum))
                self.readLengthMaximum=(readLengthMaximum if self.readLengthMaximum==None 
                                        else max(self.readLengthMaximum,readLengthMaximum))
                self.readPairedTotal+=readNumber
                self.readTotal+=readNumber
                self.totalReadLength+=totalReadLength
            self._logger.info("processed {} paired reads".format(readNumber)) 
            self.processReadsTime+=endTime-startTime
        return (readLengthMaximum,readNumber,totalReadLength,endTime-startTime)
        
        
    def _storeDirect(self, pytablesStorage):
        self.h5file["/config/"].attrs["minimumReadLength"]=self.readLengthMinimum
        self.h5file["/config/"].attrs["maximumReadLength"]=self.readLengthMaximum
        self.h5file["/config/"].attrs["numberReadsPaired"]=self.readPairedTotal
        self.h5file["/config/"].attrs["numberReadsUnpaired"]=self.readUnpairedTotal
        self.h5file["/config/"].attrs["numberReads"]=self.readTotal
        self.h5file["/config/"].attrs["totalReadLength"]=self.totalReadLength
        self.h5file["/config/"].attrs["timeProcessReads"]=int(np.ceil(self.processReadsTime))        
        
        #store merged data
        haplotyping.index.storage.Storage.storeMergedDirect(
            self.h5file, pytablesStorage, 
            self.numberOfKmers, self.minimumFrequency)

#-------------------------------------
# Main functions Indirect Connections
#-------------------------------------
        
    def _processReads(self, pytablesStorage):   
        
        if len(self.storageReadFiles)>0:
            
            #shutdown
            shutdown_event = mp.Event()
        
            #partition k-mers based on direct connections
            maxNumberOfPartitions = int(self.numberOfKmers ** (2/3))
            self.numberOfPartitions = haplotyping.index.storage.Storage.partitionKmers(self.h5file, pytablesStorage,
                                                              maxNumberOfPartitions)
            #prepare shared memory with k-mer properties and direct connections
            numberOfDirect = self.h5file["/relations/direct"].shape[0]
            shm_kmer_partition = np.dtype(haplotyping.index.Database.getUint(self.numberOfPartitions)).type
            shm_kmer_number = np.dtype(haplotyping.index.Database.getUint(self.maximumFrequency)).type
            shm_kmer_reference = np.dtype(haplotyping.index.Database.getUint(numberOfDirect)).type
            shm_kmer_numberLeft = np.dtype("uint8").type
            shm_kmer_numberRight = np.dtype("uint8").type
            shm_kmer_size = self.numberOfKmers*(shm_kmer_partition(0).nbytes+
                                                shm_kmer_number(0).nbytes+
                                                shm_kmer_reference(0).nbytes+
                                                shm_kmer_numberLeft(0).nbytes+
                                                shm_kmer_numberRight(0).nbytes)
            shm_direct_kmer = np.dtype(haplotyping.index.Database.getUint(self.numberOfKmers)).type
            shm_direct_size = numberOfDirect*shm_direct_kmer(0).nbytes
            self._logger.debug("size shared memory {} MB k-mer properties".format(math.ceil(shm_kmer_size/1048576)))
            self._logger.debug("size shared memory {} MB direct connections".format(math.ceil(shm_direct_size/1048576)))

            shm_kmer = mp.shared_memory.SharedMemory(create=True, size=shm_kmer_size)
            shm_direct = mp.shared_memory.SharedMemory(create=True, size=shm_direct_size)
            kmer_properties = np.ndarray((self.numberOfKmers,), 
                                         dtype=[("partition",shm_kmer_partition),
                                                ("number",shm_kmer_number),
                                                ("reference",shm_kmer_reference),
                                                ("numberLeft",shm_kmer_numberLeft),
                                                ("numberRight",shm_kmer_numberRight)], 
                                         buffer=shm_kmer.buf)
            direct_properties = np.ndarray((numberOfDirect,), 
                                         dtype=shm_direct_kmer, 
                                         buffer=shm_direct.buf)
            ckmerLink = 0
            for i in range(0,self.numberOfKmers,Connections.stepSizeStorage):
                #get k-mer data
                ckmers = self.h5file["/split/ckmer"][i:min(self.numberOfKmers,i+Connections.stepSizeStorage)]
                stepData = [(row[5],row[2],row[4][0],row[4][1][0],row[4][2][0]) for row in ckmers]
                kmer_properties[ckmerLink:ckmerLink+len(stepData)] = stepData
                ckmerLink+=len(stepData)
                del stepData
                #get direct data
                directStart = numberOfDirect
                directEnd = 0
                for row in ckmers:
                    if row[4][1][1]==0 and row[4][2][1]==0:
                        continue
                    else:
                        directStart=min(directStart,row[4][0])
                        directEnd=max(directEnd,row[4][0]+row[4][1][1]+row[4][2][1])
                if directStart<directEnd:
                    direct = self.h5file["/relations/direct"][directStart:directEnd]
                    stepData = [row[1][0] for row in direct]
                    direct_properties[directStart:directEnd] = stepData
                    del stepData
                                
            self._logger.debug("created shared memory {} MB k-mer properties".format(math.ceil(shm_kmer_size/1048576)))
            self._logger.debug("created shared memory {} MB direct connections".format(math.ceil(shm_direct_size/1048576)))
            
            #worker requirements
            nWorkers = max(3,mp.cpu_count()-1) if self.maximumProcesses==0 else self.maximumProcesses - 1
            nWorkersReads = min(len(self.storageReadFiles),nWorkers)            
            self._logger.debug("start {} processes to process reads".format(nWorkersReads))
            
            #queues
            queue_rawReads = mp.Queue(len(self.storageReadFiles))
            queue_filteredReads = mp.Queue()
            queue_finished = mp.Queue()            
            
            #maximum number of splitting k-mers in read
            maximumReadLength = self.h5file["/config/"].attrs["maximumReadLength"] - self.k + 1
            
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool_reads = mp.Pool(nWorkersReads, haplotyping.index.storage.Storage.workerProcessReads, 
                                 (shutdown_event,queue_rawReads,queue_filteredReads,queue_finished,
                                  self.filenameBase,self.numberOfKmers,
                                  self.numberOfPartitions,numberOfDirect,self.maximumFrequency,maximumReadLength,
                                  shm_kmer.name,shm_direct.name))
            signal.signal(signal.SIGINT, original_sigint_handler)
            
            try:
                #fill queue
                for item in self.storageReadFiles:
                    queue_rawReads.put(item)
                #trigger stopping by sending enough Nones
                for i in range(pool_reads._processes):
                    queue_rawReads.put(None)

                #now wait    
                while not (queue_rawReads.empty()):
                    time.sleep(1)

                #clean
                if not self.keepTemporaryFiles:
                   for item in self.storageReadFiles:
                       os.remove(item)
                del self.storageReadFiles

            except KeyboardInterrupt:
                self._logger.debug("caught KeyboardInterrupt")
                shutdown_event.set()
                #release memory
                shm_kmer.close()
                shm_direct.close()
                shm_kmer.unlink()
                shm_direct.unlink()
                sys.exit()
            finally:
                #terminate pool
                pool_reads.terminate()
                #close queus workers
                Connections._close_queue(queue_rawReads)
                Connections._close_queue(queue_finished)
                #shutdown
                shutdown_event.set()
                #join pool
                pool_reads.join()
                #release memory
                shm_kmer.close()
                shm_direct.close()
                try:
                    shm_kmer.unlink()
                    shm_direct.unlink()
                except Exception as e:
                    self._logger.debug("problem unlinking shared memory ({})".format(e))
                #get filtered readfiles
                storageFilteredReadFiles = Connections._collect_and_close_queue(queue_filteredReads)   
                
            #compute merged paired from storageFilteredReadFiles
            haplotyping.index.storage.Storage.combineFilteredPairs(
                storageFilteredReadFiles, pytablesStorage, self.numberOfKmers)

            #get sizes for partitions
            partitionSizes = [[0,0] for i in range(self.numberOfPartitions)]
            for item in storageFilteredReadFiles:
                with tables.open_file(item, mode="r") as pytablesStorageFiltered:
                    nReads = pytablesStorageFiltered.root.readPartitionInfo.shape[0]
                    for i in range(0,nReads,Connections.stepSizeStorage):
                        stepData = pytablesStorageFiltered.root.readPartitionInfo[i:i+Connections.stepSizeStorage]
                        for row in stepData:
                            partitionSizes[row[1]][0]+=1
                            partitionSizes[row[1]][1]+=row[0]   
            
            #shutdown
            shutdown_event = mp.Event()
        
            #worker requirements
            nWorkers = max(3,mp.cpu_count()-1) if self.maximumProcesses==0 else self.maximumProcesses - 1
            nWorkersMerges = min(len(storageFilteredReadFiles),nWorkers)     
            self._logger.debug("start {} processes to merge reads".format(nWorkersMerges))
            
            queue_ranges = mp.Queue(nWorkersMerges)
            queue_merges = mp.Queue(nWorkersMerges)
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool_merges = mp.Pool(nWorkersMerges, haplotyping.index.storage.Storage.workerMergeReads, 
                                   (shutdown_event,queue_ranges,queue_merges,
                                    storageFilteredReadFiles,partitionSizes,self.filenameBase,
                                    self.numberOfKmers,self.numberOfPartitions,maximumReadLength))
            signal.signal(signal.SIGINT, original_sigint_handler)
        
            try:
                #fill queue
                mergeNumber = math.ceil(self.numberOfPartitions/nWorkersMerges)
                for mergeStart in range(0,self.numberOfPartitions,mergeNumber):
                    queue_ranges.put((mergeStart,mergeNumber,))
                    
                #then trigger stopping by sending enough Nones
                for i in range(pool_merges._processes):
                    queue_ranges.put(None)

                #now wait    
                while not (queue_ranges.empty()):
                    time.sleep(1)

                #collect created merges
                mergeFiles = []
                while True:
                    try:
                        item = queue_merges.get(block=True, timeout=1)
                        if item==None:
                            break
                        elif isinstance(item,str):
                            mergeFiles.append(item)
                    except Empty:
                        if len(mergeFiles)==nWorkersMerges:
                            break
                        else:
                            time.sleep(1)
                #now also close this queue
                Connections._close_queue(queue_merges)

                self._logger.debug("combine {} merged files".format(len(mergeFiles)))

                #combine
                haplotyping.index.storage.Storage.combineReadMerges(
                    sorted(mergeFiles), pytablesStorage, 
                    self.numberOfKmers,self.numberOfPartitions,maximumReadLength)

                #clean mergeFiles
                if not self.keepTemporaryFiles:
                    for item in mergeFiles:
                        os.remove(item)
                        
            except KeyboardInterrupt:
                self._logger.debug("caught KeyboardInterrupt")
                shutdown_event.set()
                sys.exit()
            finally:
                #terminate pool
                pool_merges.terminate()
                #close queus workers
                Connections._close_queue(queue_ranges)
                #shutdown
                shutdown_event.set()
                #join pool
                pool_merges.join()
                
            #clean storageFilteredReadFiles
            if not self.keepTemporaryFiles:
                for item in storageFilteredReadFiles:
                    os.remove(item)
                    
    def _storeReads(self, pytablesStorage):
        #store merged data
        haplotyping.index.storage.Storage.storeMergedReads(
            self.h5file, pytablesStorage, 
            self.numberOfKmers,self.numberOfPartitions)
        
    
                
            
            
            
            



import os,sys,logging,configparser
from multiprocessing import Process,get_start_method
from flask import Flask, Blueprint, Response, render_template, current_app, g, request
from flask_restx import Api, Resource
from werkzeug.middleware.proxy_fix import ProxyFix
import json, sqlite3, re
from waitress import serve

import haplotyping

from haplotyping.service.api_tools import namespace as ns_api_tools
from haplotyping.service.api_country import namespace as ns_api_country
from haplotyping.service.api_collection import namespace as ns_api_collection
from haplotyping.service.api_variety import namespace as ns_api_variety
from haplotyping.service.api_dataset import namespace as ns_api_dataset
from haplotyping.service.api_kmer import namespace as ns_api_kmer
from haplotyping.service.api_split import namespace as ns_api_split
from haplotyping.service.api_marker import namespace as ns_api_marker

from haplotyping.service.api_kmer import cache as cache_api_kmer
from haplotyping.service.api_split import cache as cache_api_split

class API:
    
    def __init__(self, location, configFile="config.ini", doStart = True):
                                
        self.location = str(location)
        
        #set logging
        logger_server = logging.getLogger(__name__+".server")
        
        self.doStart = doStart
        self.config = configparser.ConfigParser()
        self.config.read(os.path.join(self.location,configFile)) 
        logger_server.info("read configuration file") 
        if self.config.getboolean("api","debug"):
            logger_server.info("run in debug mode") 
            logger_server.setLevel(logging.DEBUG)
        else:
            logger_server.setLevel(logging.INFO)
        
        #solve reload problem when using spawn method (osx/windows)
        if get_start_method()=="spawn":
            frame = sys._getframe()
            while frame:
                if "__name__" in frame.f_locals.keys():
                    if not frame.f_locals["__name__"]=="__main__":
                        return                    
                frame = frame.f_back
                
        #restart on errors
        while self.doStart:
            try:
                process_api = Process(target=self.process_api_messages, args=[])
                #start everything
                logger_server.info("start server on port {:s}".format(
                    self.config["api"].get("port","8080")))
                process_api.start()
                #wait until ends  
                process_api.join()
            except Exception as e:  
                logger_server.error("error: "+ str(e))  
        
    def get_db_connection():
        db_connection = getattr(g, "_database", None)
        if db_connection is None:
            config = current_app.config.get("config")
            location = current_app.config.get("location")
            db_connection = g._database = sqlite3.connect(
                os.path.join(location,config["settings"]["sqlite_db"]))
        return db_connection
    
    def get_kmc_query_library():
        try:
            config = current_app.config.get("config")
            return config["settings"]["kmc_query_library"]
        except:
            return None
    
    def get_kmc_query_binary_location():
        try:
            config = current_app.config.get("config")
            return config["settings"]["kmc_query_binary_location"]
        except:
            return None
        
    def get_data_kmer_location():
        try:
            config = current_app.config.get("config")
            return config["settings"]["data_location_kmer"]
        except:
            return None
                                
    def get_data_marker_location():
        try:
            config = current_app.config.get("config")
            return config["settings"]["data_location_marker"]
        except:
            return None
                                
    def process_api_messages(self):    
        
        #--- initialize Flask application ---  
        logging.getLogger("werkzeug").disabled = True
        #os.environ["WERKZEUG_RUN_MAIN"] = "true"
        app = Flask(__name__, static_url_path="/static", 
                    static_folder=os.path.join(self.location,"static"), 
                    template_folder=os.path.join(self.location,"templates")) 
        if self.config.getboolean("api","proxy_prefix"):
            app.wsgi_app = ProxyFix(app.wsgi_app, x_prefix=1)
        app.config["config"] = self.config
        app.config["location"] = self.location
                
        #--- blueprint ---      
        blueprint = Blueprint("api", __name__, url_prefix="/api")
        api = Api(blueprint)
        
        #logger
        logger_api = logging.getLogger(__name__)
        if self.config.getboolean("api","debug"):
            logger_api.setLevel(logging.DEBUG)
        else:
            logger_api.setLevel(logging.INFO)

        #cache
        cache_config = {
            "CACHE_TYPE": "NullCache",
            "CACHE_NO_NULL_WARNING": True
        }
        if ("cache" in app.config["config"]) and ("type" in app.config["config"]["cache"]):
            if app.config["config"]["cache"]["type"]=="SimpleCache":
                logger_api.debug("caching in memory")
                cache_config = {
                    "CACHE_TYPE": "SimpleCache"
                }
                if app.config["config"]["cache"]["timeout"]:
                    cache_config["CACHE_DEFAULT_TIMEOUT"] = int(app.config["config"]["cache"]["timeout"])
                if app.config["config"]["cache"]["threshold"]:
                    cache_config["CACHE_THRESHOLD"] = int(app.config["config"]["cache"]["threshold"])      
                cache_api_kmer.init_app(app, config=cache_config)
                cache_api_split.init_app(app, config=cache_config)
            elif (app.config["config"]["cache"]["type"]=="FileSystemCache") and ("dir" in app.config["config"]["cache"]):       
                logger_api.debug("caching on disk: "+str(app.config["config"]["cache"]["dir"]))
                cache_config = {
                    "CACHE_TYPE": "FileSystemCache",
                    "CACHE_DIR": app.config["config"]["cache"]["dir"]
                }
                if ("timeout" in app.config["config"]["cache"]) and app.config["config"]["cache"]["timeout"]:
                    cache_config["CACHE_DEFAULT_TIMEOUT"] = int(app.config["config"]["cache"]["timeout"])
                if ("threshold" in app.config["config"]["cache"]) and app.config["config"]["cache"]["threshold"]:
                    cache_config["CACHE_THRESHOLD"] = int(app.config["config"]["cache"]["threshold"]) 
                #set specific caches    
                cache_config_kmer = cache_config
                cache_config_kmer["CACHE_DIR"] = os.path.join(app.config["config"]["cache"]["dir"],"kmer")
                cache_api_kmer.init_app(app, config=cache_config_kmer)
                cache_config_split = cache_config
                cache_config_split["CACHE_DIR"] = os.path.join(app.config["config"]["cache"]["dir"],"split")
                cache_api_split.init_app(app, config=cache_config_split)                
            else:
                logger_api.debug("caching disabled")
                cache_api_kmer.init_app(app, config=cache_config)
                cache_api_split.init_app(app, config=cache_config)
        else:
            cache_api_kmer.init_app(app, config=cache_config)
            cache_api_split.init_app(app, config=cache_config)
                            
            
    
        #namespaces
        api.add_namespace(ns_api_tools)
        api.add_namespace(ns_api_country)
        api.add_namespace(ns_api_collection)
        api.add_namespace(ns_api_variety)
        api.add_namespace(ns_api_dataset)
        api.add_namespace(ns_api_kmer)
        api.add_namespace(ns_api_split)
        api.add_namespace(ns_api_marker)
        
        app.register_blueprint(blueprint) 
        app.config.SWAGGER_UI_DOC_EXPANSION = "list"

        @app.before_request
        def before_request_func():
            app.logger.info("Request to "+str(request.endpoint)+" ("+str(request.content_length)+")")
            if self.config.getboolean("api","proxy_prefix"):
                if api._schema:
                    api._schema["basePath"] = api.base_path
                    api.__schema__["basePath"] = api.base_path

        #parser = api.parser()

        #--- database ---    
        @app.teardown_appcontext
        def close_connection(exception):
            db = getattr(g, "_database", None)
            if db is not None:
                logger_api.debug("close database connection")
                db.close()

        #--- site ---
        @app.route("/")
        @app.route("/<path:path>")
        def index(path=""):
            pattern = re.compile(r"[^\/]+\/")
            rootLocation = "../"*len(re.findall(pattern, path))
            pathSplits = path.split("/")
            operation = pathSplits[0]
            if (len(pathSplits)>1):
                identifier = pathSplits[1]
            else:
                identifier = None
            variables = {
                "path": path,
                "operation": operation,
                "identifier": identifier,
                "rootLocation": rootLocation
            }
            return render_template("index.html", **variables)
        
        
        #--- start webserver ---
        if self.doStart:
            serve(app, 
                  host=self.config["api"].get("host", "::"), 
                  port=self.config["api"].get("port", "8080"), 
                  threads=self.config["api"].get("threads", "10"))   
        else:
            return app
        
        
        



import re

class General:
    
    map={"A" : "T", "C" : "G", "T" : "A", "G" : "C", "N" : "N"}
    
    kmer=re.compile(r"[ATCGN]+")
    
    def reverse_complement(kmer: str) -> str: 
        """Return the reverse complement of the provided k-mer"""
        try:
            rckmer = ""
            for base in kmer:
                rckmer += General.map[base]
            rckmer = rckmer[::-1]
            return rckmer[:]
        except Exception as e:
            raise Exception("invalid k-mer: "+str(kmer))

    def canonical(kmer: str) -> str:
        """Return the canonical form of the provided k-mer"""
        if not re.fullmatch(General.kmer, kmer):
            raise Exception("invalid k-mer: "+str(kmer))
        else:
            try:
                rckmer = ""
                do_check=True
                for (base,rcbase) in zip(kmer,kmer[::-1]):
                    rcbase = General.map[rcbase]
                    if do_check:
                        if base<rcbase:
                            return kmer
                        elif rcbase<base:
                            do_check = False
                    rckmer += rcbase
                return rckmer[:]
            except Exception as e:
                raise Exception("invalid k-mer: "+str(kmer))
        
    def is_canonical(kmer: str) -> str:
        """Test if provided k-mer is canonical"""
        try:
            for (base,rcbase) in zip(kmer,kmer[::-1]):              
                rcbase = General.map[rcbase]
                if base<rcbase:
                    return True
                elif rcbase<base:
                    return False
            return True
        except Exception as e:
            raise Exception("invalid k-mer: "+str(kmer))


import h5py, numpy as np

class Marker:
    
    def getMarkerData(filename: str, varietyId: int):
        with h5py.File(filename, mode="r") as h5file:
            varietyTable = h5file.get("/variety")
            dataTable = h5file.get("/data")
            if varietyId>=0 and varietyId<dataTable.shape[0]:
                markerList = []
                name = varietyTable["name"][varietyId].decode("utf8")
                values = dataTable[varietyId]
                for i in range(len(values)):
                    if values[i]>=0:
                        markerList.append(int(values[i]))
                    else:
                        markerList.append(None)
                return (name,markerList,)
            else:
                raise Exception("invalid identifier '"+str(varietyId)+"', data not available"+str(dataTable.shape))
                
    def getMarkerMapping(filename: str):
        with h5py.File(filename, mode="r") as h5file:
            markerTable = h5file.get("/marker")
            mappingTable = h5file.get("/mapping")
            lgTable = h5file.get("/lg")
            markerMapping = [] 
            markerNames = markerTable["name"]
            lgNames = lgTable["name"]
            mappings = {}
            for mapping in mappingTable.dtype.fields.keys():
                mappings[mapping] = [mappingTable[mapping]["lg"],mappingTable[mapping]["pos"]]
            for i in range(len(markerNames)):
                item = {
                    "name": markerNames[i].decode("utf-8"),
                    "mappings": {}
                }
                for mapping in mappings.keys():
                    lg = lgNames[mappings[mapping][0][i]].decode("utf-8")
                    pos = mappings[mapping][1][i]
                    if isinstance(pos, np.integer):
                        pos = int(pos)
                    elif isinstance(pos, np.floating):
                        pos = float(pos)
                    else:
                        pos = None
                    item["mappings"][mapping] = {"lg": lg,"pos": pos}
                markerMapping.append(item)            
            return markerMapping
        
    def getMarkerInfo(filename: str):
        with h5py.File(filename, mode="r") as h5file:
            markerTable = h5file.get("/marker")
            markerInfo = [] 
            infoData = {}
            markerNames = markerTable["name"]
            for key in markerTable.dtype.fields.keys():
                if not key=="name":
                    infoData[key] = markerTable[key]
            for i in range(len(markerNames)):
                item = {
                    "name": markerNames[i].decode("utf-8"),
                    "info": {}
                }
                for key in infoData.keys():
                    item["info"][key] = infoData[key][i].decode("utf-8")
                markerInfo.append(item)            
            return markerInfo
                
        